"""
Async DRACOON settings adapter based on httpx and pydantic
V1.2.0
(c) Octavio Simone, November 2021 

Collection of DRACOON API calls for settings (Webhooks)
Documentation: https://dracoon.team/api/swagger-ui/index.html?configUrl=/api/spec_v4/swagger-config#/groups
Webhooks documentation: https://support.dracoon.com/hc/de/articles/360013167959-Webhooks 
Please note: maximum 500 items are returned in GET requests
 - refer to documentation for details on filtering and offset
 - use documentation for payload description

"""

from typing import List
import httpx
import logging
from pydantic import validate_arguments

from dracoon.client import DRACOONClient, OAuth2ConnectionType
from dracoon.errors import InvalidArgumentError, InvalidClientError, ClientDisconnectedError
from .models import CreateWebhook, UpdateSettings, UpdateWebhook
from .responses import EventTypeList, WebhookList, Webhook, CustomerSettingsResponse

class DRACOONSettings:

    """
    API wrapper for DRACOON settings endpoint:
    Settings an webhooks management – config manager role required
    """
    
    def __init__(self, dracoon_client: DRACOONClient):
        """ requires a DRACOONClient to perform any request """
        if not isinstance(dracoon_client, DRACOONClient):
            raise InvalidClientError(message='Invalid client.')
        if dracoon_client.connection:
            self.dracoon = dracoon_client
            self.api_url = self.dracoon.base_url + self.dracoon.api_base_url + '/settings'

            self.logger = logging.getLogger('dracoon.settings')
            if self.dracoon.raise_on_err:
                self.raise_on_err = True
            else:
                self.raise_on_err = False
            self.logger.debug("DRACOON settings adapter created.")
        else:
            self.logger.error("DRACOON client error: no connection. ")
            raise ClientDisconnectedError(message='DRACOON client must be connected: client.connect()')

    @validate_arguments
    async def get_settings(self, raise_on_err: bool = False) -> CustomerSettingsResponse:
        """ list customer settings (home rooms) """
        if not await self.dracoon.test_connection() and self.dracoon.connection:
            await self.dracoon.connect(OAuth2ConnectionType.refresh_token)

        if self.raise_on_err:
            raise_on_err = True

        try:
            res = await self.dracoon.http.get(self.api_url)

            res.raise_for_status()
        except httpx.RequestError as e:
            await self.dracoon.handle_connection_error(e)
        except httpx.HTTPStatusError as e:
            self.logger.error("Getting settings failed.")
            await self.dracoon.handle_http_error(err=e, raise_on_err=raise_on_err)
        
        self.logger.info("Retrieved customer settings.")
        return CustomerSettingsResponse(**res.json())

    @validate_arguments
    async def update_settings(self, settings_update: UpdateSettings, raise_on_err: bool = False) -> CustomerSettingsResponse:
        """ update customer settings (home rooms) """
        if not await self.dracoon.test_connection() and self.dracoon.connection:
            await self.dracoon.connect(OAuth2ConnectionType.refresh_token)

        if self.raise_on_err:
            raise_on_err = True

        payload = settings_update.dict(exclude_unset=True)

        try:
            res = await self.dracoon.http.put(url=self.api_url, json=payload)

            res.raise_for_status()
        except httpx.RequestError as e:
            await self.dracoon.handle_connection_error(e)
        except httpx.HTTPStatusError as e:
            self.logger.error("Updating settings failed.")
            await self.dracoon.handle_http_error(err=e, raise_on_err=raise_on_err)
        
        self.logger.info("Updated customer settings.")
        return CustomerSettingsResponse(**res.json())
    
    def make_settings_update(self, home_rooms_active: bool = None, home_room_quota: int = None, home_room_parent_name: str = None, raise_on_err: bool = False) -> UpdateSettings:
        """ make a settings update payload required for update_settings() """
        settings_update = {}

        if home_rooms_active == False:
            raise InvalidArgumentError(message='Home rooms cannot be deactivated')

        if home_rooms_active is not None: settings_update["homeRoomsActive"] = home_rooms_active
        if home_room_quota: settings_update["homeRoomQuota"] = home_room_quota
        if home_room_parent_name: settings_update["homeRoomParentName"] = home_room_parent_name

        return UpdateSettings(**settings_update)

    @validate_arguments
    async def get_webhooks(self, offset: int = 0, filter: str = None, limit: int = None, sort: str = None, raise_on_err: bool = False) -> WebhookList:
        """ list (all) webhooks """
        if not await self.dracoon.test_connection() and self.dracoon.connection:
            await self.dracoon.connect(OAuth2ConnectionType.refresh_token)

        if self.raise_on_err:
            raise_on_err = True

        api_url = self.api_url + f'/webhooks/?offset={offset}'
        if filter != None: api_url += f'&filter={filter}' 
        if limit != None: api_url += f'&limit={str(limit)}' 
        if sort != None: api_url += f'&sort={sort}' 

        try:
            res = await self.dracoon.http.get(api_url)
            res.raise_for_status()
        except httpx.RequestError as e:
            await self.dracoon.handle_connection_error(e)
        except httpx.HTTPStatusError as e:
            self.logger.error("Getting webhooks failed.")
            await self.dracoon.handle_http_error(err=e, raise_on_err=raise_on_err)
        
        self.logger.info("Retrieved webhooks.")
        return WebhookList(**res.json())


    @validate_arguments
    async def create_webhook(self, hook: CreateWebhook, raise_on_err: bool = False) -> Webhook:
        """ creates a new webhook """
        payload = hook.dict(exclude_unset=True)

        if not await self.dracoon.test_connection() and self.dracoon.connection:
            await self.dracoon.connect(OAuth2ConnectionType.refresh_token)

        if self.raise_on_err:
            raise_on_err = True

        api_url = self.api_url + f'/webhooks'
        try:
            res = await self.dracoon.http.post(api_url, json=payload)

            res.raise_for_status()
        except httpx.RequestError as e:
            await self.dracoon.handle_connection_error(e)
        except httpx.HTTPStatusError as e:
            self.logger.error("Creating webhook failed.")
            await self.dracoon.handle_http_error(err=e, raise_on_err=raise_on_err)
        
        self.logger.info("Created webhook.")
        return Webhook(**res.json())

    def make_webhook(self, name: str, event_types: List[str], url: str, secret: str = None, 
                     is_enabled: bool = None, trigger_example: bool = None) -> CreateWebhook:
        """ make a new webhook creation payload required for create_webhook() """
        webhook = {
            "name": name,
            "eventTypeNames": event_types,
            "url": url
        }
        
        if secret: webhook["secret"] = secret
        if is_enabled is not None: webhook["isEnabled"] = is_enabled
        if trigger_example is not None: webhook["triggerExampleEvent"] = trigger_example

        return CreateWebhook(**webhook)

    def make_webhook_update(self, name: str = None, event_types: List[str] = None, url: str = None, secret: str = None, 
                     is_enabled: bool = None, trigger_example: bool = None) -> UpdateWebhook:
        """ make a new webhook update payload required for update_webhook() """
        webhook = {}
        
        if name: webhook["name"] = name
        if event_types: webhook["eventTypeNames"] = event_types
        if url: webhook["url"] = url
        if secret: webhook["secret"] = secret
        if is_enabled is not None: webhook["isEnabled"] = is_enabled
        if trigger_example is not None: webhook["triggerExampleEvent"] = trigger_example

        return UpdateWebhook(**webhook)


    # get user details for given user id
    @validate_arguments
    async def get_webhook(self, hook_id: int, raise_on_err: bool = False) -> Webhook:
        """ get webhook details for specific user (by id) """
        if not await self.dracoon.test_connection() and self.dracoon.connection:
            await self.dracoon.connect(OAuth2ConnectionType.refresh_token)

        if self.raise_on_err:
            raise_on_err = True

        api_url = self.api_url + f'/webhooks/{str(hook_id)}'

        try:
            res = await self.dracoon.http.get(api_url)

            res.raise_for_status()
        except httpx.RequestError as e:
            await self.dracoon.handle_connection_error(e)
        except httpx.HTTPStatusError as e:
            self.logger.error("Getting webhook failed.")
            await self.dracoon.handle_http_error(err=e, raise_on_err=raise_on_err)
        
        self.logger.info("Retrieved webhook.")
        return Webhook(**res.json())

    @validate_arguments
    async def update_webhook(self, hook_id: int, hook_update: UpdateWebhook, raise_on_err: bool = False) -> Webhook:

        payload = hook_update.dict(exclude_unset=True)

        if not await self.dracoon.test_connection() and self.dracoon.connection:
            await self.dracoon.connect(OAuth2ConnectionType.refresh_token)

        if self.raise_on_err:
            raise_on_err = True

        api_url = self.api_url + f'/webhooks/{str(hook_id)}'

        try:
            res = await self.dracoon.http.put(api_url, json=payload)

            res.raise_for_status()
        except httpx.RequestError as e:
            await self.dracoon.handle_connection_error(e)
        except httpx.HTTPStatusError as e:
            self.logger.error("Updating webhook failed.")
            await self.dracoon.handle_http_error(err=e, raise_on_err=raise_on_err)
        
        self.logger.info("Updated webhook.")
        return Webhook(**res.json())


    # delete user for given user id
    @validate_arguments
    async def delete_webhook(self, hook_id: int, raise_on_err: bool = False) -> None:
        if not await self.dracoon.test_connection() and self.dracoon.connection:
            await self.dracoon.connect(OAuth2ConnectionType.refresh_token)

        if self.raise_on_err:
            raise_on_err = True

        api_url = self.api_url + f'/webhooks/{str(hook_id)}'

        try:
            res = await self.dracoon.http.delete(api_url)
            res.raise_for_status()
        except httpx.RequestError as e:
            await self.dracoon.handle_connection_error(e)
        except httpx.HTTPStatusError as e:
            self.logger.error("Deleting webhook failed.")
            await self.dracoon.handle_http_error(err=e, raise_on_err=raise_on_err)
        self.logger.info("Deleted webhook.")
        return None
    
    # get user details for given user id
    @validate_arguments
    async def get_webhook_event_types(self, raise_on_err: bool = False) -> EventTypeList:

        if not await self.dracoon.test_connection() and self.dracoon.connection:
            await self.dracoon.connect(OAuth2ConnectionType.refresh_token)

        if self.raise_on_err:
            raise_on_err = True

        api_url = self.api_url + f'/webhooks/event_types'

        try:
            res = await self.dracoon.http.get(api_url)

            res.raise_for_status()
        except httpx.RequestError as e:
            await self.dracoon.handle_connection_error(e)
        except httpx.HTTPStatusError as e:
            self.logger.error("Getting webhook event types failed.")
            await self.dracoon.handle_http_error(err=e, raise_on_err=raise_on_err)
        
        self.logger.info("Retrieved webhook event types.")
        return EventTypeList(**res.json())


from pydantic import BaseModel
from typing import List, Optional
from datetime import datetime

class Expiration(BaseModel):
    enableExpiration: bool
    expireAt: datetime

# required payload for POST /groups
class CreateGroup(BaseModel):
    name: str
    expiration: Optional[Expiration]

# required payload for PUT /groups/{group_id}
class UpdateGroup(BaseModel):
    name: Optional[str]
    expiration: Optional[Expiration]



from pydantic import BaseModel
from typing import Optional, List

from dracoon.nodes.responses import Webhook
from dracoon.client.models import Range

class CustomerSettingsResponse(BaseModel):
    homeRoomsActive: bool
    homeRoomParentId: Optional[int]
    homeRoomParentName: Optional[str]
    homeRoomQuota: Optional[int]

class WebhookList(BaseModel):
    range: Range
    items: List[Webhook]

class EventType(BaseModel):
    id: int
    name: str
    usableTenantWebhook: bool
    usableCustomerAdminWebhook: bool
    usableNodeWebhook: bool
    usablePushNotification: bool

class EventTypeList(BaseModel):
    items: List[EventType]


from enum import Enum
from typing import Optional, List
from pydantic import BaseModel
from datetime import datetime

from dracoon.client.models import Range
from dracoon.nodes.models import Permissions
from dracoon.user.models import UserInfo


class AuditNodeInfo(BaseModel):
    nodeId: int
    nodeName: str
    nodeParentId: Optional[int]
    nodeIsEncrypted: Optional[bool]
    countChildren: Optional[int]


class AuditNodeInfoResponse(BaseModel):
    range: Range
    items: List[AuditNodeInfo]
    

class AuditUserPermission(BaseModel):
    userId: int
    userLogin: str
    userFirstName: str
    userLastName: str
    permissions: Permissions
    
class AuditNodeResponse(BaseModel):
    nodeId: int
    nodeName: str
    nodeParentPath: str
    nodeCntChildren: int
    auditUserPermissionList: List[AuditUserPermission]
    nodeParentId: Optional[int]
    nodeSize: Optional[int]
    nodeRecycleBinRetentionPeriod: Optional[int]
    nodeQuota: Optional[int]
    nodeIsEncrypted: Optional[bool]
    nodeHasActivitiesLog: Optional[bool]
    nodeCreatedAt: Optional[datetime]
    nodeCreatedBy: Optional[UserInfo]
    nodeUpdatedAt: Optional[datetime]
    nodeUpdatedBy: Optional[UserInfo]
    

class OperationStatus(Enum):
    Success = 0
    Error = 2

class LogEvent(BaseModel):
    id: int
    time: datetime
    userId: int
    message: str
    operationId: Optional[int]
    operationName: Optional[str]
    status: Optional[OperationStatus]
    userClient: Optional[str]
    customerId: Optional[int]
    userName: Optional[str]
    userIp: Optional[str]
    authParentSource: Optional[str]
    objectId1: Optional[int]
    objectName1: Optional[str]
    objectType1:Optional[str]
    objetcId2: Optional[int]
    objectName2:Optional[str]
    objectType2: Optional[str]
    attribute1: Optional[str]
    attribute2: Optional[str]
    attribute3: Optional[str]

class LogEventList(BaseModel):
    range: Range
    items: List[LogEvent]
    




"""
Async DRACOON reports adapter based on httpx and pydantic
V1.2.0
(c) Octavio Simone, November 2021 

Collection of DRACOON API calls for report management
Documentation: https://staging.dracoon.com/reporting/api (current beta – API on dracoon.team tbd)
Please note: maximum 500 items are returned in GET requests
 - refer to documentation for details on filtering and offset
 - use documentation for payload description

"""


from typing import List
import httpx
import logging
from pydantic import validate_arguments
from datetime import datetime


from dracoon.client import DRACOONClient, OAuth2ConnectionType
from dracoon.errors import ClientDisconnectedError, InvalidClientError
from .models import CreateReport, ReportFilter, ReportFormat, ReportSubType, ReportType
from .responses import ReportList


class DRACOONReports:

    """
    API wrapper for DRACOON reports API:
    Reports management - requires auditor role.
    """

    def __init__(self, dracoon_client: DRACOONClient):
        """ requires a DRACOONClient to perform any request """
        if not isinstance(dracoon_client, DRACOONClient):
            raise InvalidClientError(message='Invalid client')
        if dracoon_client.connection:
            self.dracoon = dracoon_client
            self.api_url = self.dracoon.base_url + self.dracoon.reporting_base_url + '/reports'
            self.logger = logging.getLogger('dracoon.reports')
            if self.dracoon.raise_on_err:
                self.raise_on_err = True
            else:
                self.raise_on_err = False
            self.logger.debug("DRACOON reports adapter created.")
        else:
            self.logger.error("DRACOON client error: no connection. ")
            raise ClientDisconnectedError(message='DRACOON client must be connected: client.connect()')

    @validate_arguments
    async def create_report(self, report: CreateReport, raise_on_err: bool = False) -> None:
        """ create a new report """
        payload = report.dict(exclude_unset=True)

        if not await self.dracoon.test_connection() and self.dracoon.connection:
            await self.dracoon.connect(OAuth2ConnectionType.refresh_token)

        if self.raise_on_err:
            raise_on_err = True

        try:
            res = await self.dracoon.http.post(self.api_url, json=payload)

            res.raise_for_status()
        except httpx.RequestError as e:
            await self.dracoon.handle_connection_error(e)
        except httpx.HTTPStatusError as e:
            self.logger.error("Creating report failed.")
            await self.dracoon.handle_http_error(err=e, raise_on_err=raise_on_err)
        
        self.logger.info("Created report.")
        return None
    
    def make_report(self, name: str, target_id: int, formats: List[ReportFormat], type: ReportType = ReportType.single, 
                    sub_type: ReportSubType = ReportSubType.audit_report, 
                    enabled: bool = None, filter: ReportFilter = None) -> CreateReport:
        """ make a report payload to generate a report """
        report = {
            "name": name,
            "target": {
                "id": target_id
            },
            "type": type,
            "subType": sub_type,
            "formats": formats  
        }

        if enabled is not None: report["enabled"] = enabled
        if filter: report["filter"] = filter

        return CreateReport(**report)


    def make_report_filter(self, from_date: datetime = None, to_date: datetime = None, parent_room_id: int = None, 
                           user_id: int = None, operations: List[int] = None) -> ReportFilter:
        """ make an optional report filter needed for make_report() """

        filter = {}

        if from_date: filter["fromDate"] = from_date
        if to_date: filter["toDate"] = to_date
        if parent_room_id: filter["parentRoom"] = parent_room_id
        if user_id: filter["userId"] = user_id
        if operations: filter["operations"] = operations

        return ReportFilter(**filter)
    
    @validate_arguments
    async def get_reports(self, name: str = None, type: str = None, sub_type: str = None, state: str = None, 
                has_error: bool = None, enabled: bool = None, 
                offset: int = 0, limit: int = None, sort: str = None, raise_on_err: bool = False) -> ReportList:
        """ list (all) reports """
        if not await self.dracoon.test_connection() and self.dracoon.connection:
            await self.dracoon.connect(OAuth2ConnectionType.refresh_token)

        if self.raise_on_err:
            raise_on_err = True

        api_url = self.api_url + f'/?offset={offset}'
        if name: api_url += f'&name={name}' 
        if limit != None: api_url += f'&limit={str(limit)}' 
        if sort != None: api_url += f'&sort={sort}' 
        if type != None: api_url += f'&type={type}' 
        if sub_type != None: api_url += f'&subType={sub_type}'
        if enabled != None: api_url += f'&type={str(enabled).lower()}' 
        if has_error != None: api_url += f'&hasError={str(has_error).lower()}'
        if state != None: api_url += f'&state={state}'

        try:
            res = await self.dracoon.http.get(self.api_url)

            res.raise_for_status()
        except httpx.RequestError as e:
            await self.dracoon.handle_connection_error(e)
        except httpx.HTTPStatusError as e:
            self.logger.error("Getting reports failed.")
            await self.dracoon.handle_http_error(err=e, raise_on_err=raise_on_err)
        
        self.logger.info("Retrieved reports.")
        return ReportList(**res.json())

    @validate_arguments
    async def delete_reports(self, report_list: List[int], raise_on_err: bool = False) -> None:
        """ delete a list of reports (by ids) """
        if not await self.dracoon.test_connection() and self.dracoon.connection:
            await self.dracoon.connect(OAuth2ConnectionType.refresh_token)

        if self.raise_on_err:
            raise_on_err = True

        payload = {
            "ids": report_list
        }
        
        try:
            res = await self.dracoon.http.request(method='DELETE', url=self.api_url, json=payload, headers=self.dracoon.http.headers)
            res.raise_for_status()
        except httpx.RequestError as e:
            await self.dracoon.handle_connection_error(e)
        except httpx.HTTPStatusError as e:
            self.logger.error("Deleting reports failed.")
            await self.dracoon.handle_http_error(err=e, raise_on_err=raise_on_err)

        self.logger.info("Deleted reports.")
        return None

    @validate_arguments
    async def delete_report(self, report_id: int, raise_on_err: bool = False) -> None:
        """ delete a specific report (by id) """
        if not await self.dracoon.test_connection() and self.dracoon.connection:
            await self.dracoon.connect(OAuth2ConnectionType.refresh_token)

        if self.raise_on_err:
            raise_on_err = True

        api_url = self.api_url + f'/{str(report_id)}'

        try:
            res = await self.dracoon.http.delete(api_url)
            res.raise_for_status()
        except httpx.RequestError as e:
            await self.dracoon.handle_connection_error(e)
        except httpx.HTTPStatusError as e:
            self.logger.error("Deleting report failed.")
            await self.dracoon.handle_http_error(err=e, raise_on_err=raise_on_err)
        
        self.logger.info("Deleted report.")
        return None



from dracoon.crypto.models import UserKeyPairContainer, UserKeyPairVersion
from dracoon.user import DRACOONUser
from dracoon import DRACOONClient, OAuth2ConnectionType
import dotenv
import os
import asyncio

from dracoon.user.models import UpdateAccount, UserAccount


dotenv.load_dotenv()

client_id = os.environ.get('CLIENT_ID')
client_secret = os.environ.get('CLIENT_SECRET')
username = os.environ.get('E2E_USER_NAME')
password = os.environ.get('E2E_PASSWORD')
base_url = os.environ.get('E2E_BASE_URL')
base_url_server = os.environ.get('E2E_SERVER_BASE_URL')

async def test_user_e2e():

    dracoon = DRACOONClient(base_url=base_url, client_id=client_id, client_secret=client_secret)
    await dracoon.connect(OAuth2ConnectionType.password_flow, username=username, password=password)
    user = DRACOONUser(dracoon_client=dracoon)
    assert isinstance(user, DRACOONUser)
    print('Connection test complete (/)')
        
    user_info = await user.get_account_information(more_info=True)
    assert isinstance(user_info, UserAccount)
    print('Getting user info passed (/)')

    account_update = user.make_account_update(phone='999999999')
    assert isinstance(account_update, UpdateAccount)
    print('Making account update payload test passed (/)')

    account_update_res = await user.update_account_information(account_update=account_update)

    assert account_update_res.phone == '999999999'
    print('Updating user account test passed (/)')

    secret = 'VerySecret123!'

    keypair_2048 = await user.set_user_keypair(secret=secret, version=UserKeyPairVersion.RSA2048)
    assert keypair_2048 == None

    print('Setting keypair RSA 2048 passed (/)')

    keypair_2048 = await user.get_user_keypair()

    assert isinstance(keypair_2048, UserKeyPairContainer)
    assert keypair_2048.privateKeyContainer.version == UserKeyPairVersion.RSA2048.value
    assert keypair_2048.publicKeyContainer.version == UserKeyPairVersion.RSA2048.value

    print('Getting keypair RSA 2048 passed (/)')

    del_keypair = await user.delete_user_keypair(version=UserKeyPairVersion.RSA2048)
    assert del_keypair == None

    print('Deleting keypair test passed (/)')

    keypair_4096 = await user.set_user_keypair(secret=secret, version=UserKeyPairVersion.RSA4096)
    assert keypair_4096 == None

    print('Setting keypair RSA 4096 passed (/)')

    keypair_4096 = await user.get_user_keypair()

    assert isinstance(keypair_4096, UserKeyPairContainer)
    assert keypair_4096.privateKeyContainer.version == UserKeyPairVersion.RSA4096.value
    assert keypair_4096.publicKeyContainer.version == UserKeyPairVersion.RSA4096.value

    print('Getting keypair RSA 4096 passed (/)')

    del_keypair = await user.delete_user_keypair(version=UserKeyPairVersion.RSA4096)
    assert del_keypair == None

    print('User E2E tests passed (/)')
    
async def test_user_e2e_server():

    dracoon = DRACOONClient(base_url=base_url_server, client_id=client_id, client_secret=client_secret)
    await dracoon.connect(OAuth2ConnectionType.password_flow, username=username, password=password)
    user = DRACOONUser(dracoon_client=dracoon)
    assert isinstance(user, DRACOONUser)
    print('Connection test complete (/)')
        
    user_info = await user.get_account_information(more_info=True)
    assert isinstance(user_info, UserAccount)
    print('Getting user info passed (/)')

    account_update = user.make_account_update(phone='999999999')
    assert isinstance(account_update, UpdateAccount)
    print('Making account update payload test passed (/)')

    account_update_res = await user.update_account_information(account_update=account_update)

    assert account_update_res.phone == '999999999'
    print('Updating user account test passed (/)')

    secret = 'VerySecret123!'

    keypair_2048 = await user.set_user_keypair(secret=secret, version=UserKeyPairVersion.RSA2048)
    assert keypair_2048 == None

    print('Setting keypair RSA 2048 passed (/)')

    keypair_2048 = await user.get_user_keypair()

    assert isinstance(keypair_2048, UserKeyPairContainer)
    assert keypair_2048.privateKeyContainer.version == UserKeyPairVersion.RSA2048.value
    assert keypair_2048.publicKeyContainer.version == UserKeyPairVersion.RSA2048.value

    print('Getting keypair RSA 2048 passed (/)')

    del_keypair = await user.delete_user_keypair(version=UserKeyPairVersion.RSA2048)
    assert del_keypair == None

    print('Deleting keypair test passed (/)')

    keypair_4096 = await user.set_user_keypair(secret=secret, version=UserKeyPairVersion.RSA4096)
    assert keypair_4096 == None

    print('Setting keypair RSA 4096 passed (/)')

    keypair_4096 = await user.get_user_keypair()

    assert isinstance(keypair_4096, UserKeyPairContainer)
    assert keypair_4096.privateKeyContainer.version == UserKeyPairVersion.RSA4096.value
    assert keypair_4096.publicKeyContainer.version == UserKeyPairVersion.RSA4096.value

    print('Getting keypair RSA 4096 passed (/)')

    del_keypair = await user.delete_user_keypair(version=UserKeyPairVersion.RSA4096)
    assert del_keypair == None

    print('User E2E tests passed (/)')


if __name__ == '__main__':
    asyncio.run(test_user_e2e())
    asyncio.run(test_user_e2e_server())

    

   





""" 
Core models required by other models 

"""


from pydantic import BaseModel
from typing import Dict, Optional, Union
from datetime import datetime


class Range(BaseModel):
    offset: int
    limit: int
    total: int

class Expiration(BaseModel):
    enableExpiration: bool
    expireAt: datetime

class ErrorMessage(BaseModel):
    code: int
    message: str
    debugInfo: Optional[str]
    errorCode: Optional[int]

# adheres to proxy model from httpx (dict with http / https and respective str) or single str
ProxyConfig = Union[Dict[str, str], str]


import unittest
import sys
from dracoon import crypto
from dracoon.crypto.models import FileKey, FileKeyVersion, PlainFileKey, PlainFileKeyVersion, PlainUserKeyPairContainer, UserKeyPairContainer, UserKeyPairVersion

class TestDRACOONCrypto(unittest.TestCase):
    """ Test cases for the DRACOON Python Crypto package """

    def test_plain_user_keypair_generation_2048(self):
        """ tests function to create a plain RSA 2048 keypair """
        plain_keypair = crypto.create_plain_userkeypair(version=UserKeyPairVersion.RSA2048)
        self.assertIsInstance(plain_keypair, PlainUserKeyPairContainer)
        self.assertEqual(plain_keypair.privateKeyContainer.version, UserKeyPairVersion.RSA2048.value)
        self.assertEqual(plain_keypair.publicKeyContainer.version, UserKeyPairVersion.RSA2048.value)

    def test_plain_user_keypair_generation_4096(self):
        """ tests function to create a plain RSA 4096 keypair """
        plain_keypair = crypto.create_plain_userkeypair(version=UserKeyPairVersion.RSA4096)
        self.assertIsInstance(plain_keypair, PlainUserKeyPairContainer)
        self.assertEqual(plain_keypair.privateKeyContainer.version, UserKeyPairVersion.RSA4096.value)
        self.assertEqual(plain_keypair.publicKeyContainer.version, UserKeyPairVersion.RSA4096.value)

    def test_private_key_encryption(self):

        """ Test en- and decryption of a private key with given secret """

        secret = 'VerySecret1234!'
        plain_2048_keypair = crypto.create_plain_userkeypair(version=UserKeyPairVersion.RSA2048)
        plain_4096_keypair = crypto.create_plain_userkeypair(version=UserKeyPairVersion.RSA4096)
        encrypted_2048_keypair = crypto.encrypt_private_key(secret=secret, plain_key=plain_2048_keypair)
        encrypted_4096_keypair = crypto.encrypt_private_key(secret=secret, plain_key=plain_4096_keypair)

        self.assertIsInstance(encrypted_2048_keypair, UserKeyPairContainer)
        self.assertIsInstance(encrypted_4096_keypair, UserKeyPairContainer)
        self.assertNotEqual(plain_2048_keypair, encrypted_2048_keypair)
        self.assertNotEqual(plain_4096_keypair, encrypted_4096_keypair)

        decrypted_2048_keypair = crypto.decrypt_private_key(secret=secret, keypair=encrypted_2048_keypair)
        decrypted_4096_keypair = crypto.decrypt_private_key(secret=secret, keypair=encrypted_4096_keypair)

        self.assertEqual(plain_2048_keypair, decrypted_2048_keypair)
        self.assertEqual(plain_4096_keypair, decrypted_4096_keypair)


    def test_encryption_on_the_fly(self):
        """ Test en- and decryption on the fly with given plain file keys """

        plain_2048_file_key = crypto.create_file_key(version=PlainFileKeyVersion.AES256GCM)
        plain_4096_file_key = crypto.create_file_key(version=PlainFileKeyVersion.AES256GCM)

        message = b'Very secret message'

        message_enc_2048, plain_key_2048 = crypto.encrypt_bytes(plain_data=message, plain_file_key=plain_2048_file_key)
        message_enc_4096, plain_key_4096 = crypto.encrypt_bytes(plain_data=message, plain_file_key=plain_4096_file_key)

        """ Message no longer readable """
        self.assertNotEqual(message_enc_2048, message)
        self.assertNotEqual(message_enc_4096, message)
        
        """ After finalizing, the key contains a tag """
        self.assertIsNotNone(plain_key_2048.tag)
        self.assertIsNotNone(plain_key_4096.tag)
        
        """ Decrypting the message with obtained file key should deliver the same message """
        decrypted_2048 = crypto.decrypt_bytes(enc_data=message_enc_2048, plain_file_key=plain_key_2048)
        decrypted_4096 = crypto.decrypt_bytes(enc_data=message_enc_4096, plain_file_key=plain_key_4096)

        """ Plain keys should adhere to format """
        self.assertIsInstance(plain_key_2048, PlainFileKey)
        self.assertIsInstance(plain_key_4096, PlainFileKey)
        
        """ Plain keys be equal, as tag has been added after encryption – received key is returned and updated """
        self.assertEqual(plain_2048_file_key, plain_key_2048)
        self.assertEqual(plain_4096_file_key, plain_key_4096)
        
        """ Ensure original message is received """
        self.assertEqual(message, decrypted_2048)
        self.assertEqual(message, decrypted_4096)

    def test_chunked_encryption(self):
        """ Test chunked en- and decryption with given plain file keys """

        plain_2048_file_key = crypto.create_file_key(version=PlainFileKeyVersion.AES256GCM)
        plain_4096_file_key = crypto.create_file_key(version=PlainFileKeyVersion.AES256GCM)

        message = 'Very secret message'

        encryptor_2048 = crypto.FileEncryptionCipher(plain_file_key=plain_2048_file_key)
        encryptor_4096 = crypto.FileEncryptionCipher(plain_file_key=plain_4096_file_key)

        message_enc_2048 = b''
        message_enc_4096 = b''

        """ Encrypt each character """
        for char in message:
            buf = char.encode()
            message_enc_2048 += encryptor_2048.encode_bytes(plain_data=buf)
            message_enc_4096 += encryptor_4096.encode_bytes(plain_data=buf)

        """ Finalize encryption """
        last_data_2048, plain_2048_key = encryptor_2048.finalize()
        last_data_4096, plain_4096_key = encryptor_4096.finalize()

        message_enc_2048 += last_data_2048
        message_enc_4096 += last_data_4096
         
        """ Ensure message is no longer readable """
        self.assertNotEqual(message_enc_2048, message)
        self.assertNotEqual(message_enc_4096, message)
        
        decryptor_2048 = crypto.FileDecryptionCipher(plain_file_key=plain_2048_key)
        decryptor_4096 = crypto.FileDecryptionCipher(plain_file_key=plain_4096_key)

        message_2048 = b''
        message_4096 = b''


        for chunk in message_enc_2048:
            chunk = chunk.to_bytes(1, sys.byteorder)      
            message_2048 += decryptor_2048.decode_bytes(chunk)
            
        for chunk in message_enc_4096:
            chunk = chunk.to_bytes(1, sys.byteorder)      
            message_4096 += decryptor_4096.decode_bytes(chunk)
        
        
        message_2048 += decryptor_2048.finalize()
        message_4096 += decryptor_4096.finalize()

        decrypted_2048_fly = crypto.decrypt_bytes(enc_data=message_enc_2048, plain_file_key=plain_2048_key)
        decrypted_4096_fly = crypto.decrypt_bytes(enc_data=message_enc_4096, plain_file_key=plain_4096_key)

        self.assertEqual(message_2048, decrypted_2048_fly)
        self.assertEqual(message_4096, decrypted_4096_fly)

        self.assertEqual(message.encode(), message_2048)
        self.assertEqual(message.encode(), message_4096)


    def test_file_key_encryption(self):
        """ Test encryption of a plain file key for both versions, ensure version check correct """
        plain_file_key_2048 = crypto.create_file_key(PlainFileKeyVersion.AES256GCM)
        plain_file_key_4096 = crypto.create_file_key(PlainFileKeyVersion.AES256GCM)
        plain_keypair_2048 = crypto.create_plain_userkeypair(UserKeyPairVersion.RSA2048)
        plain_keypair_4096 = crypto.create_plain_userkeypair(UserKeyPairVersion.RSA4096)

        enc_file_key_2048 = crypto.encrypt_file_key(plain_file_key=plain_file_key_2048, keypair=plain_keypair_2048)
        enc_file_key_4096 = crypto.encrypt_file_key(plain_file_key=plain_file_key_4096, keypair=plain_keypair_4096)

        self.assertIsInstance(enc_file_key_2048, FileKey)
        self.assertIsInstance(enc_file_key_4096, FileKey)

        self.assertEqual(enc_file_key_2048.version, FileKeyVersion.RSA2048_AES256GCM.value)
        self.assertEqual(enc_file_key_4096.version, FileKeyVersion.RSA_4096_AES256GCM.value)

        decrypted_key_2048 = crypto.decrypt_file_key(file_key=enc_file_key_2048, keypair=plain_keypair_2048)
        decrypted_key_4096 = crypto.decrypt_file_key(file_key=enc_file_key_4096, keypair=plain_keypair_4096)

        self.assertEqual(plain_file_key_2048, decrypted_key_2048)
        self.assertEqual(plain_file_key_4096, decrypted_key_4096)

        self.assertIsInstance(decrypted_key_2048, PlainFileKey)
        self.assertIsInstance(decrypted_key_4096, PlainFileKey)
        
    def test_public_file_key_encrytion(self):
        """ Test encryption of a plain file key for both versions using a public key """
        plain_file_key_2048 = crypto.create_file_key(PlainFileKeyVersion.AES256GCM)
        plain_file_key_4096 = crypto.create_file_key(PlainFileKeyVersion.AES256GCM)
        
        plain_keypair_2048 = crypto.create_plain_userkeypair(UserKeyPairVersion.RSA2048)
        plain_keypair_4096 = crypto.create_plain_userkeypair(UserKeyPairVersion.RSA4096)

        enc_file_key_2048 = crypto.encrypt_file_key_public(public_key=plain_keypair_2048.publicKeyContainer, plain_file_key=plain_file_key_2048)
        enc_file_key_4096 = crypto.encrypt_file_key_public(public_key=plain_keypair_4096.publicKeyContainer, plain_file_key=plain_file_key_4096)

        self.assertIsInstance(enc_file_key_2048, FileKey)
        self.assertIsInstance(enc_file_key_4096, FileKey)

        self.assertEqual(enc_file_key_2048.version, FileKeyVersion.RSA2048_AES256GCM.value)
        self.assertEqual(enc_file_key_4096.version, FileKeyVersion.RSA_4096_AES256GCM.value)

        decrypted_key_2048 = crypto.decrypt_file_key(file_key=enc_file_key_2048, keypair=plain_keypair_2048)
        decrypted_key_4096 = crypto.decrypt_file_key(file_key=enc_file_key_4096, keypair=plain_keypair_4096)

        self.assertEqual(plain_file_key_2048, decrypted_key_2048)
        self.assertEqual(plain_file_key_4096, decrypted_key_4096)

        self.assertIsInstance(decrypted_key_2048, PlainFileKey)
        self.assertIsInstance(decrypted_key_4096, PlainFileKey)
        

if __name__ == '__main__':
    unittest.main()


from pydantic import BaseModel
from typing import List, Optional

# required payload for PUT /settings
class UpdateSettings(BaseModel):
    homeRoomParentName: Optional[str]
    homeRoomQuota: Optional[int]
    homeRoomsActive: Optional[bool]

# required payload for POST /settings/webhooks
class CreateWebhook(BaseModel):
    name: str
    eventTypeNames: List[str]
    url: str
    secret: Optional[str]
    isEnabled: Optional[bool]
    triggerExampleEvent: Optional[bool]

# required payload for PUT /settings/webhooks/{webhook_id}
class UpdateWebhook(BaseModel):
    name: Optional[str]
    eventTypeNames: Optional[List[str]]
    url: Optional[str]
    secret: Optional[str]
    isEnabled: Optional[bool]
    triggerExampleEvent: Optional[bool]








from pydantic import BaseModel
from typing import List, Optional
from datetime import datetime, time
from enum import Enum

class ReportType(Enum):
    single = 'single'
    periodic = 'periodic'

class ReportSubType(Enum):
    audit_report = 'general-audit-report'
    user_activity = 'user-audit-report'

class ReportFormat(Enum):
    pdf = 'pdf'
    csv = 'csv-plain'
    csv_semicolon = 'csv-semicolon-delimited'

class ReportFilter(BaseModel):
    fromDate: Optional[datetime]
    toDate: Optional[datetime]
    parentRoom: Optional[int]
    userId: Optional[int]
    operations: Optional[List[int]]

class ReportExecutionType(Enum):
    on_demand = 'on-demand'
    monthly = 'monthly'

class ReportExecution(BaseModel):
    type: ReportExecutionType
    timeOfDay: time
    dayOfMonth: int

class ReportTarget(BaseModel):
    id: int

class CreateReport(BaseModel):
    name: str
    type: ReportType
    subType: ReportSubType
    enabled: Optional[bool]
    execution: Optional[ReportExecution]
    formats: List[ReportFormat]
    filter: Optional[ReportFilter]
    target: ReportTarget






"""
DRACOON Crypto utils based on cryptography
V1.2.0

(c) Octavio Simone, November 2021 

Collection of DRACOON API calls
Documentation: https://dracoon.team/api

Please note: maximum 500 items are returned in GET requests 

 - refer to documentation for details on filtering and offset 
 - use documentation for payload description 
 
Encryption: 
 - Files are encrypted with AES 256 GCM cipher
 - Keypair (public and private key) is RSA-4096 (2048 supported as well)

"""
import os
import base64
import logging 
from typing import Tuple

from pydantic import validate_arguments
from cryptography.hazmat.primitives.asymmetric import rsa, padding
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.ciphers import (Cipher, algorithms, modes)


from dracoon.errors import (InvalidKeypairVersionError, FileKeyEncryptionError, CryptoMissingDataError, InvalidFileKeyError)
from .models import (FileKey, FileKeyVersion, PlainFileKey, PlainFileKeyVersion,  
                                   PlainUserKeyPairContainer, PublicKeyContainer, UserKeyPairContainer, 
                                   UserKeyPairVersion)

logger = logging.getLogger('dracoon.crypto')

@validate_arguments
def encrypt_private_key(secret: str, plain_key: PlainUserKeyPairContainer) -> UserKeyPairContainer:
    """ encrypt a private key (requires a plain user keypair container: create_plain_user_keypair()) """
    logger.info("Encrypting private key - version: %s", plain_key.privateKeyContainer.version)
    # load plain private key
    private_key_pem = plain_key.privateKeyContainer.privateKey
    private_key: rsa.RSAPrivateKeyWithSerialization = serialization.load_pem_private_key(
        data=private_key_pem.encode('ascii'), password=None)

    # serialize key with passphrase (secret)
    encrypted_private_key = private_key.private_bytes(encoding=serialization.Encoding.PEM,
                                                      format=serialization.PrivateFormat.PKCS8,
                                                      encryption_algorithm=serialization.BestAvailableEncryption(secret.encode('ascii')))

    return UserKeyPairContainer(**{

        "privateKeyContainer": {
            "version": plain_key.privateKeyContainer.version,
            "privateKey": encrypted_private_key.decode('ascii')
        },
        "publicKeyContainer": {
            "version": plain_key.publicKeyContainer.version,
            "publicKey": plain_key.publicKeyContainer.publicKey
        }
    })

@validate_arguments
def decrypt_private_key(secret: str, keypair: UserKeyPairContainer) -> PlainUserKeyPairContainer:
    """ decrypt a private key (requires secret). Returns a plain user keypair """
    logger.info("Decrypting private key - version: %s", keypair.privateKeyContainer.version)
    # load encrypted private key
    private_key_pem = keypair.privateKeyContainer.privateKey
    private_key: rsa.RSAPrivateKeyWithSerialization = serialization.load_pem_private_key(
        data=private_key_pem.encode('ascii'), password=secret.encode('ascii'), )

    # export without passphrase
    private_key_pem = private_key.private_bytes(encoding=serialization.Encoding.PEM,
                                                format=serialization.PrivateFormat.TraditionalOpenSSL,
                                                encryption_algorithm=serialization.NoEncryption())

    return PlainUserKeyPairContainer(**{

        "privateKeyContainer": {
            "version": keypair.privateKeyContainer.version,
            "privateKey": private_key_pem.decode('ascii')
        },
        "publicKeyContainer": {
            "version": keypair.publicKeyContainer.version,
            "publicKey": keypair.publicKeyContainer.publicKey
        }
    })


@validate_arguments
def create_plain_userkeypair(version: UserKeyPairVersion) -> PlainUserKeyPairContainer:
    """ create a new RSA plain keypair – needs to be encrypted with encrypt_private_key() """
    
    logger.info("Creating plain user keypair - version: %s", version.value)

    # generate asymmetric RSA key based on version
    if version == UserKeyPairVersion.RSA2048:
        key = rsa.generate_private_key(public_exponent=65537, key_size=2048)
    elif version == UserKeyPairVersion.RSA4096:
        key = rsa.generate_private_key(public_exponent=65537, key_size=4096)
    else:
        raise InvalidKeypairVersionError(message='Invalid keypair version')
    private_key_pem = key.private_bytes(encoding=serialization.Encoding.PEM,
                                        format=serialization.PrivateFormat.TraditionalOpenSSL,
                                        encryption_algorithm=serialization.NoEncryption())
    public_key = key.public_key()
    public_key_pem = public_key.public_bytes(
        encoding=serialization.Encoding.PEM, format=serialization.PublicFormat.SubjectPublicKeyInfo)

    return PlainUserKeyPairContainer(**{

        "privateKeyContainer": {
            "version": version.value,
            "privateKey": private_key_pem.decode('ascii')
        },
        "publicKeyContainer": {
            "version": version.value,
            "publicKey": public_key_pem.decode('ascii')
        }
    })


def get_file_key_version(keypair: PlainUserKeyPairContainer) -> FileKeyVersion:
    """ get required file key version from given user keypair """

    logger.info("Getting file key version: %s", keypair.privateKeyContainer.version)

    if keypair.publicKeyContainer.version == UserKeyPairVersion.RSA2048.value and keypair.privateKeyContainer.version == UserKeyPairVersion.RSA2048.value:
        return FileKeyVersion.RSA2048_AES256GCM
    elif keypair.publicKeyContainer.version == UserKeyPairVersion.RSA4096.value and keypair.privateKeyContainer.version == UserKeyPairVersion.RSA4096.value:
        return FileKeyVersion.RSA_4096_AES256GCM
    else:
        raise InvalidKeypairVersionError(message='Invalid keypair version')


def get_file_key_version_public(public_key: PublicKeyContainer) -> FileKeyVersion:
    """ get file required file key version from given public key (needed for missing file keys request) """
    
    logger.info("Getting file key version: %s", public_key.version)

    if public_key.version == UserKeyPairVersion.RSA2048.value:
        return FileKeyVersion.RSA2048_AES256GCM
    elif public_key.version == UserKeyPairVersion.RSA4096.value:
        return FileKeyVersion.RSA_4096_AES256GCM
    else:
        raise InvalidKeypairVersionError(message='Invalid keypair version')


@validate_arguments
def create_file_key(version: PlainFileKeyVersion = PlainFileKeyVersion.AES256GCM) -> PlainFileKey:
    """ create a plain file key (AES 256) """

    logger.info("Creating file key: %s", version.value)

    # get random bytes for key and vector
    key = os.urandom(32)
    iv = os.urandom(12)

    # base64 encode
    encoded_key = base64.b64encode(key)
    encoded_iv = base64.b64encode(iv)

    return PlainFileKey(**{
        "version": version.value,
        "key": encoded_key.decode('ascii'),
        "iv": encoded_iv.decode('ascii'),
        "tag": None
    })


def encrypt_file_key_public(plain_file_key: PlainFileKey, public_key: PublicKeyContainer) -> FileKey:
    """ encrypt a file key with given public key container """

    logger.info("Creating file key with public key: %s", public_key.version)

    public_key_pem = serialization.load_pem_public_key(
        public_key.publicKey.encode('ascii'))
        
    # check correct version
    file_key_version = get_file_key_version_public(public_key)

    key = plain_file_key.key

    # initialize variable for encrypted file key
    encrypted_key = None

    # use SHA1 MGF1 and SHA256 hash
    if public_key.version == UserKeyPairVersion.RSA2048.value:
        encrypted_key = public_key_pem.encrypt(plaintext=base64.b64decode(key), padding=padding.OAEP(mgf=padding.MGF1(algorithm=hashes.SHA256()),
                                                                                                     algorithm=hashes.SHA1(),
                                                                                                     label=None))
    # use SHA256 (hash and MGF1)
    elif public_key.version == UserKeyPairVersion.RSA4096.value:
        encrypted_key = public_key_pem.encrypt(plaintext=base64.b64decode(key), padding=padding.OAEP(mgf=padding.MGF1(algorithm=hashes.SHA256()),
                                                                                                     algorithm=hashes.SHA256(),
                                                                                                     label=None))
    return FileKey(**{
        "version": file_key_version.value,
        "key": base64.b64encode(encrypted_key).decode(),
        "iv": plain_file_key.iv,
        "tag": plain_file_key.tag
    })

# encrypt a plain file key


def encrypt_file_key(plain_file_key: PlainFileKey, keypair: PlainUserKeyPairContainer) -> FileKey: 
    """ encrypt a file key with given plain user keypair """

    logger.info("Encrypting file key: %s", keypair.privateKeyContainer.version)

    private_key_pem = keypair.privateKeyContainer.privateKey
    private_key: rsa.RSAPrivateKeyWithSerialization = serialization.load_pem_private_key(
        data=private_key_pem.encode('ascii'), password=None)
    public_key = private_key.public_key()

    # check correct version
    file_key_version = get_file_key_version(keypair)

    key = plain_file_key.key

    # initialize variable for encrypted file key
    encrypted_key = None

    # use SHA1 MGF1 and SHA256 hash
    if keypair.publicKeyContainer.version == UserKeyPairVersion.RSA2048.value:
        encrypted_key = public_key.encrypt(plaintext=base64.b64decode(key), padding=padding.OAEP(mgf=padding.MGF1(algorithm=hashes.SHA256()),
                                                                                                 algorithm=hashes.SHA1(),
                                                                                                 label=None))
    # use SHA256 (hash and MGF1)
    elif keypair.publicKeyContainer.version == UserKeyPairVersion.RSA4096.value:
        encrypted_key = public_key.encrypt(plaintext=base64.b64decode(key), padding=padding.OAEP(mgf=padding.MGF1(algorithm=hashes.SHA256()),
                                                                                                 algorithm=hashes.SHA256(),
                                                                                                 label=None))
    return FileKey(**{
        "version": file_key_version.value,
        "key": base64.b64encode(encrypted_key).decode(),
        "iv": plain_file_key.iv,
        "tag": plain_file_key.tag
    })


def decrypt_file_key(file_key: FileKey, keypair: PlainUserKeyPairContainer) -> PlainFileKey:
    """ decrypt a file key with given plain user keypair """

    logger.info("Decrypting file key: %s", keypair.privateKeyContainer.version)

    key = base64.b64decode(file_key.key)
    private_key_pem = keypair.privateKeyContainer.privateKey
    private_key: rsa.RSAPrivateKeyWithSerialization = serialization.load_pem_private_key(
        data=private_key_pem.encode('ascii'), password=None)

    file_key_version = get_file_key_version(keypair)

    plain_key = None

    if keypair.publicKeyContainer.version == UserKeyPairVersion.RSA2048.value:
        plain_key = private_key.decrypt(ciphertext=key, padding=padding.OAEP(mgf=padding.MGF1(algorithm=hashes.SHA256()),
                                                                                  algorithm=hashes.SHA1(),
                                                                                  label=None))
    # use SHA256 (hash and MGF1)
    elif keypair.publicKeyContainer.version == UserKeyPairVersion.RSA4096.value:
        plain_key = private_key.decrypt(ciphertext=key, padding=padding.OAEP(mgf=padding.MGF1(algorithm=hashes.SHA256()),
                                                                                  algorithm=hashes.SHA256(),
                                                                                  label=None))

    if plain_key != None:
        return PlainFileKey(**{
            "version": PlainFileKeyVersion.AES256GCM.value,
            "key": base64.b64encode(plain_key).decode('ascii'),
            "iv": file_key.iv,
            "tag": file_key.tag
        })
    else:
        raise FileKeyEncryptionError(message='Could not encrypt file key')


def decrypt_bytes(enc_data: bytes, plain_file_key: PlainFileKey) -> bytes:
    """ decrypt bytes with given plain file key (on the fly) """
    logger.info("Decrypting bytes with version: %s", plain_file_key.version)

    if enc_data == None:
        raise CryptoMissingDataError(message='No data to process.')

    key = base64.b64decode(plain_file_key.key)
    iv = base64.b64decode(plain_file_key.iv)
    tag = base64.b64decode(plain_file_key.tag)

    decryptor = Cipher(algorithm=algorithms.AES(
        key), mode=modes.GCM(iv, tag)).decryptor()

    plain_bytes = decryptor.update(data=enc_data) + decryptor.finalize()

    return plain_bytes


def encrypt_bytes(plain_data: bytes, plain_file_key: PlainFileKey) -> Tuple[bytes, PlainFileKey]:
    """ encrypt bytes with given plain file key (on the fly) """
    logger.info("Encrypting bytes with version: %s", plain_file_key.version)

    if not bytes:
        raise CryptoMissingDataError(message='No data to process.')

    key = base64.b64decode(plain_file_key.key)
    iv = base64.b64decode(plain_file_key.iv)

    encryptor = Cipher(algorithm=algorithms.AES(key),
                       mode=modes.GCM(iv)).encryptor()
    enc_bytes = encryptor.update(plain_data) + encryptor.finalize()

    plain_file_key.tag = base64.b64encode(encryptor.tag).decode('ascii')

    return (enc_bytes, plain_file_key)


class FileEncryptionCipher:
    """
    Instantiates an encryption cipher
    To be used for chunking
    Initialized with plain file key

    Usage: 
    cipher = FileEncryptionCipher(plain_file_key)

    with open(some_file, 'rb') as plain_file:
        while True:
            chunk = plain_file.read(CHUNK_SIZE)
            if not chunk: break

            cipher.encode_bytes(chunk)

        enc_bytes, plain_file_key = cipher.finalize()

    """

    def __init__(self, plain_file_key: PlainFileKey):
        """ initialize encryptor with a plain file key """
        
        if not plain_file_key.version or not plain_file_key.key or not plain_file_key.iv:
            logger.critical("Invalid file key format (lack of iv or key).")
            raise InvalidFileKeyError(message="Invalid file key")
        if plain_file_key.version != PlainFileKeyVersion.AES256GCM.value and plain_file_key.version != PlainFileKeyVersion.AES256GCM.value:
            logger.critical("Invalid / unknown file key version.")
            raise InvalidFileKeyError(message="Invalid file key version")
        
        logger.info("Initialized encryptor with version: %s", plain_file_key.version)

        self.plain_file_key = plain_file_key

        self.key = base64.b64decode(plain_file_key.key)
        self.iv = base64.b64decode(plain_file_key.iv)

        self.encryptor = Cipher(algorithm=algorithms.AES(
            self.key), mode=modes.GCM(self.iv)).encryptor()

    def encode_bytes(self, plain_data: bytes) -> bytes:
        """ encode bytes """
        logger.debug("Encrypting bytes...")
        return self.encryptor.update(plain_data)

    def finalize(self) -> Tuple[bytes, PlainFileKey]:
        """ complete encryption """
        logger.debug("Finalizing encryption...")
        enc_bytes = self.encryptor.finalize()
        self.plain_file_key.tag = base64.b64encode(
            self.encryptor.tag).decode('ascii')

        return enc_bytes, self.plain_file_key


class FileDecryptionCipher:
    """
    Instantiates a decryption cipher
    To be used for chunking
    Initialized with plain file key

    Usage: 
    cipher = FileDecryptionCipher(plain_file_key)

    with open(some_file, 'rb') as encoded_file:
        while True:
            chunk = encoded_file.read(CHUNK_SIZE)
            if not chunk: break

            cipher.decode_bytes(chunk)

        plain_bytes = cipher.finalize()

    """

    def __init__(self, plain_file_key: PlainFileKey):
        """ initialize decryptor with a plain file key """
        if not plain_file_key.version or not plain_file_key.key or not plain_file_key.iv:
            logger.critical("Invalid file key format (lack of iv or key).")
            raise InvalidFileKeyError(message="Invalid file key")
        if plain_file_key.version != PlainFileKeyVersion.AES256GCM.value and plain_file_key.version != PlainFileKeyVersion.AES256GCM.value:
            logger.critical("Invalid / unknown file key version.")
            raise InvalidFileKeyError(message="Invalid file key version")
        self.plain_file_key = plain_file_key

        self.key = base64.b64decode(plain_file_key.key)
        self.iv = base64.b64decode(plain_file_key.iv)
        self.tag = base64.b64decode(plain_file_key.tag)

        self.encryptor = Cipher(algorithm=algorithms.AES(
            self.key), mode=modes.GCM(self.iv, self.tag)).decryptor()

        logger.info("Initialized encryptor with version: %s", plain_file_key.version)

    def decode_bytes(self, enc_data: bytes) -> bytes:
        """ decode bytes """
        logger.debug("Decrypting bytes...")
        return self.encryptor.update(enc_data)

    def finalize(self) -> bytes:
        """ complete decryption """
        logger.debug("Finalizing encryption...")
        plain_bytes = self.encryptor.finalize()

        return plain_bytes



from pydantic import BaseModel
from typing import List, Optional
from datetime import datetime
from dracoon.crypto.models import UserKeyPairContainer, FileKey

class Expiration(BaseModel):
    enableExpiration: bool
    expireAt: datetime

# required payload for POST /shares/downloads
class CreateShare(BaseModel):
    nodeId: int
    name: Optional[str]
    password: Optional[str]
    expiration: Optional[Expiration]
    notes: Optional[str]
    internalNotes: Optional[str]
    showCreatorName: Optional[bool]
    showCreatorUsername: Optional[bool]
    maxDownloads: Optional[int]
    keyPair: Optional[UserKeyPairContainer]
    fileKey: Optional[FileKey]
    reciverLanguage: Optional[str]
    textMessageRecipients: Optional[List[str]]

# required payload for PUT /shares/downloads
class UpdateShares(BaseModel):
    expiration: Optional[Expiration]
    showCreatorName: Optional[bool]
    showCreatorUsername: Optional[bool]
    maxDownloads: Optional[int]
    resetMaxDownloads: Optional[bool]
    objectIds: List[int]

# required payload for PUT /shares/downloads/{share_id}
class UpdateShare(BaseModel):
    expiration: Optional[Expiration]
    showCreatorName: Optional[bool]
    showCreatorUsername: Optional[bool]
    maxDownloads: Optional[int]
    resetMaxDownloads: Optional[bool]
    name: Optional[str]
    password: Optional[str]
    notes: Optional[str]
    internalNotes: Optional[str]
    defaultCountry: Optional[str]
    resetPassword: Optional[bool]

# required payload for POST /shares/downloads/{share_id}/email and /shares/loads/{share_id}/email 
class SendShare(BaseModel):
    recipients: List[str]
    body: str
    receiverLanguage: Optional[str]

# required payload for POST /shares/uploads
class CreateFileRequest(BaseModel):
    targetId: int
    name: Optional[str]
    password: Optional[str]
    expiration: Optional[Expiration]
    filesExpiryPeriod: Optional[int]
    notes: Optional[str]
    internalNotes: Optional[str]
    showUploadedFiles: Optional[bool]
    maxSlots: Optional[int]
    maxSize: Optional[int]
    receiverLanguage: Optional[str]
    textMessageRecipients: Optional[List[str]]
    showCreatorName: Optional[bool]
    showCreatorUsername: Optional[bool]

# required payload for PUT /shares/uploads
class UpdateFileRequests(BaseModel):
    expiration: Optional[Expiration]
    showCreatorName: Optional[bool]
    showCreatorUsername: Optional[bool]
    showUploadedFiles: Optional[bool]
    maxSlots: Optional[int]
    maxSize: Optional[int]
    filesExpiryPeriod: Optional[int]
    resetFilesExpiryPeriod: Optional[bool]
    objectIds: List[int]

# required payload for PUT /shares/uploads/{share_id}
class UpdateFileRequest(BaseModel):
    name: Optional[str]
    password: Optional[str]
    expiration: Optional[Expiration]
    notes: Optional[str]
    internalNotes: Optional[str]
    showCreatorName: Optional[bool]
    showCreatorUsername: Optional[bool]
    showUploadedFiles: Optional[bool]
    maxSlots: Optional[int]
    maxSize: Optional[int]
    filesExpiryPeriod: Optional[int]
    resetFilesExpiryPeriod: Optional[bool]
    resetMaxSize: Optional[bool]
    resetMaxSlots: Optional[bool]
    textMessageRecipients: Optional[List[str]]
    defaultCountry: Optional[str]
    resetPassword: Optional[bool]
    receiverLanguage: Optional[str]




    









"""
Async DRACOON client based on httpx
V1.2.0

(c) Octavio Simone, November 2021 

The client implements all login and logout procedures and is part of every API adapter.

"""

import base64 
from enum import Enum
from dataclasses import dataclass
import base64
import asyncio
from datetime import datetime
import logging
import re

import httpx
from dracoon.client.models import ProxyConfig

from dracoon.errors import (MissingCredentialsError, HTTPBadRequestError, HTTPUnauthorizedError, 
                            HTTPPaymentRequiredError, HTTPForbiddenError, HTTPNotFoundError, HTTPConflictError, HTTPPreconditionsFailedError,
                            HTTPUnknownError)

USER_AGENT = 'dracoon-python-1.5.1'
DEFAULT_TIMEOUT_CONFIG = httpx.Timeout(10, connect=20, read=20)

class OAuth2ConnectionType(Enum):
    """ enum as connection type for DRACOONClient """
    """ supports authorization code flow, password flow and refresh token """
    password_flow = 1
    auth_code = 2
    refresh_token = 3

@dataclass
class DRACOONConnection:
    """ DRACOON connection with tokens and validity """
    connected_at: datetime
    access_token: str
    access_token_validity: int
    refresh_token: str



class DRACOONClient:
    """ DRACOON client with an httpx async client """
    """ requires OAuth connection details and base url """
    api_base_url = '/api/v4'
    branding_base_url = '/branding/api'
    reporting_base_url = '/reporting/api'
    headers = {
        "User-Agent": USER_AGENT

    }

    def __init__(self, base_url: str, client_id: str = 'dracoon_legacy_scripting', client_secret: str = '', raise_on_err: bool = False,
                 proxy_config: ProxyConfig = None):
        """ client is initialized with DRACOON instance details (url and OAuth client credentials) """
        self.base_url = base_url
        self.client_id = client_id
        self.client_secret = client_secret
        self.http = httpx.AsyncClient(headers=self.headers, timeout=DEFAULT_TIMEOUT_CONFIG, proxies=proxy_config)
        self.uploader = httpx.AsyncClient(timeout=DEFAULT_TIMEOUT_CONFIG, proxies=proxy_config)
        self.downloader = httpx.AsyncClient(timeout=DEFAULT_TIMEOUT_CONFIG, proxies=proxy_config)
        self.connected = False
        self.connection: DRACOONConnection = None
        self.raise_on_err = raise_on_err
        self.logger = logging.getLogger('dracoon.client')

        self.logger.info("DRACOON client created.")
        self.logger.debug(f"DRACOON client config: {self.base_url} // {self.client_id}")

    def __del__(self):
        """ on client destroy terminate async client """

        # handle asyncio runtime
        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            loop = None
        
        # close httpx client 
        if loop and loop.is_running():
            loop.create_task(self.http.aclose())
            loop.create_task(self.uploader.aclose())
            loop.create_task(self.downloader.aclose())
        elif loop and not loop.is_running():
            loop.run_until_complete(self.http.aclose())
            loop.run_until_complete(self.downloader.aclose())
            loop.run_until_complete(self.uploader.aclose())
            

   
    async def connect(self, connection_type: OAuth2ConnectionType, username: str = None, password: str = None, 
                      auth_code: str = None, refresh_token: str = None, redirect_uri: str = None) -> DRACOONConnection:
        """ connects based on given OAuth2ConnectionType """
        token_url = self.base_url + '/oauth/token'
        now = datetime.now()

        self.logger.debug("Establishing DRACOON connection...")
        
        # handle missing credentials for password flow
        if connection_type == OAuth2ConnectionType.password_flow and username == None and password == None:
            await self.http.aclose()
            self.logger.error("Missing credentials for OAuth2 password flow.")
            raise MissingCredentialsError(message='Username and password are mandatory for OAuth2 password flow.')

        # handle missing auth code for authorization code flow 
        if connection_type == OAuth2ConnectionType.auth_code and auth_code == None:
            await self.http.aclose()
            self.logger.error("Missing auth code for OAuth2 password flow.")
            raise MissingCredentialsError(message='Auth code is mandatory for OAuth2 authorization code flow.')
                
        # get connection via OAuth2 password flow
        if connection_type == OAuth2ConnectionType.password_flow:
            data = {'grant_type': 'password',
                'username': username, 'password': password}
            
            token_payload = base64.b64encode(
                bytes(self.client_id + ':' + self.client_secret, 'ascii'))

            self.http.headers["Authorization"] = "Basic " + \
                    token_payload.decode('ascii')
            try:
                res = await self.http.post(url=token_url, data=data)
                self.logger.debug("Request status code %s", res.status_code)
                res.raise_for_status()
            except httpx.RequestError as e:
                await self.handle_connection_error(e)

            except httpx.HTTPStatusError as e:
                self.logger.debug("Login error: %s", e.response.text)
                self.logger.error("Password flow authentication failed.")
                await self.handle_http_error(e, True)


            self.connection = DRACOONConnection(now, res.json()["access_token"], res.json()["expires_in"],
                                         res.json()["refresh_token"])


        if connection_type == OAuth2ConnectionType.auth_code:
            
            
            if redirect_uri:
                _redirect_uri = redirect_uri
            else:
                _redirect_uri = self.base_url + '/oauth/callback'

            data = {'grant_type': 'authorization_code', 'code': auth_code, 'client_id': self.client_id, 'client_secret': self.client_secret, 'redirect_uri': _redirect_uri}

            try:
                res = await self.http.post(url=token_url, data=data)
                res.raise_for_status()
            except httpx.RequestError as e:
                await self.handle_connection_error(e)

            except httpx.HTTPStatusError as e:
                self.logger.debug("Login error: %s", e.response.text)
                self.logger.error("Authorization code authentication failed.")
                await self.handle_http_error(e, True)

            self.connection = DRACOONConnection(now, res.json()["access_token"], res.json()["expires_in"],
                                         res.json()["refresh_token"])

            self.logger.info("Established connection.")
            self.logger.debug("Access token valid: %s", self.connection.access_token_validity)



        if connection_type == OAuth2ConnectionType.refresh_token:

            if self.connection:
                refresh_token = self.connection.refresh_token
            if refresh_token == None:
                raise MissingCredentialsError(message='Missing refresh token.')

            data = {'grant_type': 'refresh_token', 'refresh_token': refresh_token, 'client_id': self.client_id, 'client_secret': self.client_secret}


            try:
                res = await self.http.post(url=token_url, data=data)
                res.raise_for_status()
            except httpx.RequestError as e:
                await self.handle_connection_error(e)

            except httpx.HTTPStatusError as e:
                self.logger.debug("Login error: %s", e.response.text)
                self.logger.error("Refresh token authentication failed.")
                await self.handle_http_error(e, True)


            self.connection = DRACOONConnection(now, res.json()["access_token"], res.json()["expires_in"],
                                         res.json()["refresh_token"])

        self.connected = True
        self.http.headers["Authorization"] = "Bearer " + self.connection.access_token
  

        return self.connection

    async def disconnect(self):
        await self.http.aclose()

    def get_code_url(self):
        """ builds OAuth authorization code url to visit – requires OAuth app to use redirect uri ($host/oauth/callback) """
        # generate URL string for OAuth auth code flow
        return self.base_url + f'/oauth/authorize?branding=full&response_type=code&client_id={self.client_id}&redirect_uri={self.base_url}/oauth/callback&scope=all'


    async def logout(self, revoke_refresh_token: bool = False):
        """ revoke tokens """
        revoke_url = self.base_url + '/oauth/revoke'
    
        access_data = {'token': self.connection.access_token, 'token_type_hint': 'access_token', 'client_id': self.client_id, 'client_secret': self.client_secret}
        refresh_data = {'token': self.connection.refresh_token, 'token_type_hint': 'refresh_token', 'client_id': self.client_id, 'client_secret': self.client_secret}

        try:
            res_a = await self.http.post(url=revoke_url, data=access_data)
            res_a.raise_for_status()
            if revoke_refresh_token:
                res_r = await self.http.post(url=revoke_url, data=refresh_data)
                res_r.raise_for_status()
        except httpx.RequestError as e:
            await self.handle_connection_error(e)
        except httpx.HTTPStatusError as e:
            self.logger.debug("Token revoke error: %s", e.response.text)
            self.logger.error("Revoking token(s) failed.")
            await self.handle_http_error(err=e, raise_on_err=self.raise_on_err)

        self.connected = False
        self.connection = None
        await self.http.aclose()
        await self.downloader.aclose()
        await self.uploader.aclose()

    
    async def check_access_token(self, test: bool = False):
        """ check access token validity (based on connection time and token validity) """
        self.logger.debug("Testing access token validity.")
        self.logger.debug("Full authenticated test enabled: %s", test)

        if not test and self.connection:
            now = datetime.now()
            return (now - self.connection.connected_at).seconds < self.connection.access_token_validity
        elif test and self.connection:
            return await self.test_connection()
        else:
            self.logger.error("Access token no longer valid.")
            return False

    async def test_connection(self, test: bool = False) -> bool:
        """ test authenticated connection via authenticated ping """
        self.logger.debug("Testing authenticated connection.")

        if not self.connection or not self.connected:
            self.logger.critical("DRACOON connection not established.")
            return False

        if test:

            test_url = self.base_url + '/api/v4/user/ping'

            try:
                res = await self.http.get(url=test_url)
                res.raise_for_status()

            except httpx.RequestError as e:
                await self.handle_connection_error(e)

            except httpx.HTTPStatusError as e:
                self.logger.error("Authenticated ping failed.")
                await self.handle_http_error(err=e, raise_on_err=self.raise_on_err)
                return False
            
            return True

        return await self.check_access_token()


    async def handle_http_error(self, err: httpx.HTTPStatusError, raise_on_err: bool, is_xml: bool = False, close_client: bool = False):
        if self.raise_on_err:
            raise_on_err = self.raise_on_err
        
        self.logger.error("HTTP request failed: %s", err.response.status_code)
        
        # handle AWS S3 XML responses
        if not is_xml:
            self.logger.debug("%s", err.response.text)
        elif is_xml:
            self.logger.debug("%s", err.response.content)

        if close_client:
            await self.http.aclose()
        
        if raise_on_err:
            self.raise_http_error(err=err)

    async def handle_connection_error(self, err: httpx.RequestError):
        self.logger.critical("Connection error.")
        self.logger.critical(err.request.url)
        await self.http.aclose()
        raise err
    
    async def handle_generic_error(self, err: Exception, close_client: bool = False):
        self.logger.critical("An error ocurred.")
        self.logger.debug("%s", err)
        if close_client:
            await self.http.aclose()
        raise err

    def raise_http_error(self, err: httpx.HTTPStatusError):

        if err.response.status_code == 400:
            raise HTTPBadRequestError(error=err)
        if err.response.status_code == 401:
            raise HTTPUnauthorizedError(error=err)
        if err.response.status_code == 402:
            raise HTTPPaymentRequiredError(error=err)
        if err.response.status_code == 403:
            raise HTTPForbiddenError(error=err)
        if err.response.status_code == 404:
            raise HTTPNotFoundError(error=err)
        if err.response.status_code == 409:
            raise HTTPConflictError(error=err)
        if err.response.status_code == 412:
            raise HTTPPreconditionsFailedError(error=err)
        else:
            raise HTTPUnknownError(error=err)

        
        


"""
Async DRACOON downloads adapter based on httpx, tqdm and pydantic
V1.2.0
(c) Octavio Simone, February 2022 

Documentation: https://dracoon.team/api/swagger-ui/index.html?configUrl=/api/spec_v4/swagger-config#/downloads
Please note: maximum 500 items are returned in GET requests
 - refer to documentation on how to upload files:
https://support.dracoon.com/hc/de/articles/115005512089

"""
import os
from pathlib import Path
import logging

from tqdm import tqdm
import httpx

from dracoon.nodes import CHUNK_SIZE
from dracoon.nodes.models import Callback, Node, NodeType
from dracoon.client import DRACOONClient
from dracoon.crypto import FileDecryptionCipher, decrypt_file_key
from dracoon.crypto.models import FileKey, PlainUserKeyPairContainer
from dracoon.errors import (InvalidClientError, ClientDisconnectedError, InvalidFileError, 
                            FileConflictError, InvalidPathError)


class DRACOONDownloads:

    """
    API wrapper for DRACOON uploads endpoint:
    Upload files via proxy upload with given open upload channel.
    """

    def __init__(self, dracoon_client: DRACOONClient):
        """ requires a DRACOONClient to perform any request """
        if not isinstance(dracoon_client, DRACOONClient):
            raise InvalidClientError(message='Invalid client.')
        if dracoon_client.connection:
            self.dracoon = dracoon_client
            self.api_url = self.dracoon.base_url + self.dracoon.api_base_url + '/user'
            self.logger = logging.getLogger('dracoon.downloads')
            if self.dracoon.raise_on_err:
                self.raise_on_err = True
            else:
                self.raise_on_err = False

            self.logger.debug("DRACOON downloads adapter created.")
   
        else:
            self.dracoon.logger.critical("DRACOON client not connected.")
            raise ClientDisconnectedError(message='DRACOON client must be connected: client.connect()')

    
    def check_file_exists(self, path: str) -> bool:
        """ Check if a file already exists. """

        file = Path(path)

        return file.exists() and file.is_file()

    
    async def download_unencrypted(self, download_url: str, target_path: str, node_info: Node, chunksize: int = CHUNK_SIZE, 
                                   raise_on_err: bool = False, display_progress: bool = False,
                                   callback_fn: Callback  = None,
                                   file_name: str = None
                                   ):
        """ Download a file from an unecrypted data room. """

        self.logger.info("Download started.")
        self.logger.debug("Download to %s", target_path)

        if self.raise_on_err:
            raise_on_err = True
        
        if node_info.type != NodeType.file:
            self.logger.critical("Invalid node type: %s", node_info.type)
            err = InvalidFileError(message='Ony file download possible.')
            await self.dracoon.handle_generic_error(err=err)
        
        if file_name is None:
            file_path = os.path.join(target_path, node_info.name)
        elif file_name is not None:
            file_path = os.path.join(target_path, file_name)
            
        if self.check_file_exists(file_path):
            self.logger.critical("File already exists: %s", file_path)
            err = FileConflictError('File already exists.')
            await self.dracoon.handle_generic_error(err=err)

        folder = Path(target_path)

        if not folder.is_dir():
            self.logger.critical("Target path is not a folder.")
            err = InvalidPathError(f'A folder needs to be provided. {target_path} is not a folder.')
            await self.dracoon.handle_generic_error(err=err)
        
        size = node_info.size
        
        # init callback size
        if callback_fn: callback_fn(0, size)

        self.logger.debug("File download for size: %s", size)
        self.logger.debug("Using chunksize: %s", chunksize)
            
        try:             
            file_out = open(file_path, 'wb')
                
            async with self.dracoon.downloader.stream(method='GET', url=download_url) as res:
                    
                if display_progress: progress = tqdm(unit='iMB',unit_divisor=1024, total=size, unit_scale=True, desc=node_info.name)
                async for chunk in res.aiter_bytes(chunksize):
                    file_out.write(chunk)
                    if display_progress: progress.update(len(chunk))
                    if callback_fn: callback_fn(len(chunk))
                                        
        except httpx.RequestError as e:
            await self.dracoon.handle_connection_error(e)
        except httpx.HTTPStatusError as e:
            await self.dracoon.handle_http_error(err=e, raise_on_err=raise_on_err)
        finally:
            if file_out: file_out.close()
            if display_progress and progress: progress.close()
                   
        self.logger.info("Download completed.")
            

    async def download_encrypted(self, download_url: str, target_path: str, node_info: Node, plain_keypair: PlainUserKeyPairContainer, file_key: FileKey, 
                                       chunksize: int = CHUNK_SIZE, raise_on_err: bool = False, display_progress: bool = False, 
                                       callback_fn: Callback  = None, file_name: str = None):   
        """ Download a file from an encrypted data room. """

        self.logger.info("Download started.")
        self.logger.debug("Download to %s", target_path)

        if self.raise_on_err:
            raise_on_err = True
            
        if not node_info:
            err = InvalidFileError(message='File does not exist.')
            await self.dracoon.handle_generic_error(err=err)
        
        if node_info.type != NodeType.file:
            err = InvalidFileError(message='Ony file download possible.')
            await self.dracoon.handle_generic_error(err=err)

        if file_name is None:
            file_path = os.path.join(target_path, node_info.name)
        elif file_name is not None:
            file_path = os.path.join(target_path, file_name)

        if self.check_file_exists(file_path):
            await self.dracoon.logout()
            self.logger.critical("File already exists: %s", file_path)
            err = FileConflictError(message='File already exists.')
            await self.dracoon.handle_generic_error(err=err)

        folder = Path(target_path)

        if not folder.is_dir():
            await self.dracoon.logout()
            self.logger.critical("Target path is not a folder.")
            err = InvalidPathError(f'A file needs to be provided. {target_path} is not a folder.')
            await self.dracoon.handle_generic_error(err=err)
        
        size = node_info.size

        plain_file_key = decrypt_file_key(file_key=file_key, keypair=plain_keypair)
        decryptor = FileDecryptionCipher(plain_file_key=plain_file_key)

        self.logger.debug("File download for size: %s", size)
        self.logger.debug("Using chunksize: %s", chunksize)
        
        # init callback size
        if callback_fn: callback_fn(0, size)

        try:         
            file_out = open(file_path, 'wb')  
                
            async with self.dracoon.downloader.stream(method='GET', url=download_url) as res:
                res.raise_for_status()
                if display_progress: progress = tqdm(unit='iMB',unit_divisor=1024, total=size, unit_scale=True, desc=node_info.name)   
                    
                # decrypt file and then write to disk
                async for chunk in res.aiter_bytes(chunksize):            
                    plain_chunk = decryptor.decode_bytes(chunk)
                    file_out.write(plain_chunk)
                    if display_progress: progress.update(len(chunk))
                    if callback_fn: callback_fn(len(chunk))
                        
                        # finalize encryption after last chunk
                    if not chunk:
                        last_data = decryptor.finalize()
                        file_out.write(last_data)
                                     
                    self.logger.info("Download completed.")
        except httpx.RequestError as e:
            await self.dracoon.handle_connection_error(e)
        except httpx.HTTPStatusError as e:
            await self.dracoon.handle_http_error(err=e, raise_on_err=raise_on_err)
        finally:
            if file_out: file_out.close()
            if display_progress and progress: progress.close()




                
            



from pydantic import BaseModel
from typing import Optional, List
from dracoon.crypto.models import FileKey

class UploadChannelResponse(BaseModel):
    uploadUrl: str
    uploadId: str
    token: str

class UserFileKey(BaseModel):
    userId: int
    fileKey: FileKey

class UserFileKeyList(BaseModel):
    items: List[UserFileKey]

class FinalizeUpload(BaseModel):
    resolutionStrategy: Optional[str]
    keepShareLinks: Optional[bool]
    fileName: Optional[str]
    fileKey: Optional[FileKey]
    userFileKeyList: Optional[UserFileKeyList]

    



from logging import Logger, StreamHandler, FileHandler, Formatter, getLogger

def create_logger(log_level: int, log_stream: bool, log_file_out: bool = False, log_file: str = None) -> Logger:

    logger = getLogger('dracoon')

    formatter = Formatter('%(asctime)s %(name)-12s %(levelname)-8s %(message)s')

    if log_stream:
        stream_handler = StreamHandler()
        stream_handler.setFormatter(formatter)
        logger.addHandler(stream_handler)
        
    if log_file_out and log_file is not None:
        file_handler = FileHandler(filename=log_file)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)
        
    logger.setLevel(log_level)
    
    return logger



from enum import Enum
from typing_extensions import Protocol
from pydantic import BaseModel
from typing import Any, List, Optional
from datetime import datetime


from dracoon.user.models import UserInfo
from dracoon.crypto.models import EncryptionInfo, FileKey, PublicKeyContainer
from dracoon.client.models import Range

# required payload for POST /nodes/{node_id}/comments
class CommentNode(BaseModel):
    text: str

class NodeItem(BaseModel):
    id: int
    name: Optional[str]
    timestampCreation: Optional[datetime]
    timestampModification: Optional[datetime]

# required payload for POST /nodes/{node_id}/copy_to or /node/{node_id}/move_to
class TransferNode(BaseModel):
    items: List[NodeItem]
    resolutionStrategy: Optional[str]
    keepShareLinks: Optional[bool]

# required payload for POST /nodes/deleted_nodes/actions/restore
class RestoreNode(BaseModel):
    deletedNodeIds: List[int]
    resolutionStrategy: Optional[str]
    keepShareLinks: Optional[bool]
    parentId: Optional[int]

class SetFileKeysItem(BaseModel):
    fileId: int
    userId: int
    fileKey: FileKey

# required payload for POST /nodes/files/keys
class SetFileKeys(BaseModel):
    items: List[SetFileKeysItem]

class Expiration(BaseModel):
    enableExpiration: bool
    expireAt: str

# required payload for POST /nodes/files/uploads
class CreateUploadChannel(BaseModel):
    parentId: int
    name: str
    classification: Optional[int]
    size: Optional[int]
    expiration: Optional[Expiration]
    notes: Optional[str]
    directS3Upload: Optional[bool]
    timestampCreation: Optional[str]
    timestampModification: Optional[str]


class S3Part(BaseModel):
    partNumber: int
    partEtag: str

class CompleteS3Upload(BaseModel):
    parts: List[S3Part]
    resolutionStrategy: Optional[str]
    keepShareLinks: Optional[bool]
    fileName: Optional[str]
    fileKey: Optional[FileKey]


class CompleteUpload(BaseModel):
    resolutionStrategy: Optional[str]
    keepShareLinks: Optional[bool]
    fileName: Optional[str]
    fileKey: Optional[FileKey]
    
class GetS3Urls(BaseModel):
    firstPartNumber: int
    lastPartNumber: int
    size: int

# required payload for POST /nodes/folders
class CreateFolder(BaseModel):
    parentId: int
    name: str
    notes: Optional[str]
    timestampCreation: Optional[datetime]
    timestampModificiation: Optional[datetime]

# required payload for PUT /nodes/folders/{folder_id}
class UpdateFolder(BaseModel):
    name: Optional[str]
    notes: Optional[str]
    timestampCreation: Optional[str]
    timestampModificiation: Optional[str]

# required payload for POST /nodes/rooms
class CreateRoom(BaseModel):
    parentId: Optional[int]
    name: str
    notes: Optional[str]
    quota: Optional[int]
    recycleBinRetentionPeriod: Optional[int]
    inheritPermissions: Optional[bool]
    hasActivitiesLog: Optional[bool]
    newGroupMemberAcceptance: Optional[str]
    adminGroupIds: Optional[List[int]]
    adminIds: Optional[List[int]]
    classification: Optional[int]
    timestampCreation: Optional[str]
    timestampModificiation: Optional[str]

# required payload for PUT /nodes/rooms/{room_id}
class UpdateRoom(BaseModel):
    name: Optional[str]
    quota: Optional[int]
    notes: Optional[str]
    timestampCreation: Optional[str]
    timestampModificiation: Optional[str]

# required payload for PUT /nodes/rooms/{room_id}/config
class ConfigRoom(BaseModel):
    adminGroupIds: Optional[List[int]]
    adminIds: Optional[List[int]]
    recycleBinRetentionPeriod: Optional[int]
    inheritPermissions: Optional[bool]
    takeOverPermissions: Optional[bool]
    hasActivitiesLog: Optional[bool]
    newGroupMemberAcceptance: Optional[str]
    classification: Optional[int]

# required payload for PUT /nodes/files
class UpdateFiles(BaseModel):
    classification: Optional[int]
    expiration: Optional[Expiration]
    objectIds: List[int]

class UpdateFile(BaseModel):
    name: Optional[str]
    expiration: Optional[Expiration]
    classification: Optional[int]
    notes: Optional[str]
    timestampCreation: Optional[str]
    timestampModification: Optional[str]


class Permissions(BaseModel):
    manage: bool
    read: bool
    create: bool
    change: bool
    delete: bool
    manageDownloadShare: bool
    manageUploadShare: bool
    readRecycleBin: bool
    restoreRecycleBin: bool
    deleteRecycleBin: bool


class UpdateRoomUserItem(BaseModel):
    id: int
    permissions: Permissions

class UpdateRoomGroupItem(UpdateRoomUserItem):
    id: int
    permissions: Permissions
    newGroupMemberAcceptance: Optional[str]

# required payload for PUT /nodes/rooms/{room_id}/groups
class UpdateRoomGroups(BaseModel):
    items: List[UpdateRoomGroupItem]

# required payload for PUT /nodes/rooms/{room_id}/users
class UpdateRoomUsers(BaseModel):
    items: List[UpdateRoomUserItem]

class UpdateRoomHookItem(BaseModel):
    webhookId: int
    isAssigned: bool

# required payload for PUT /nodes/rooms/{room_id}/webhooks
class UpdateRoomHooks(BaseModel):
    items: List[UpdateRoomHookItem]


class ProcessRoomPendingItem(BaseModel):
    userId: int
    groupId: int
    roomId: int
    roomName: str
    state: str

# required payload for PUT /nodes/rooms/{room_id}/pending
class ProcessRoomPendingUsers(BaseModel):
    items: List[ProcessRoomPendingItem]


class NodeType(Enum):
    file = "file"
    folder = "folder"
    room = "room"

class Node(BaseModel):
    id: int
    type: NodeType
    referenceId: Optional[int]
    name: str
    timestampCreation: Optional[datetime]
    timestampModification: Optional[datetime]
    parentId: Optional[int]
    parentPath: Optional[str]
    createdAt: Optional[datetime]
    updatedAt: Optional[datetime]
    createdBy: Optional[UserInfo]
    updatedBy: Optional[UserInfo]
    expireAt: Optional[datetime]
    hash: Optional[str]
    fileType: Optional[str]
    mediaType: Optional[str]
    size: Optional[int]
    classification: Optional[int]
    notes: Optional[str]
    permissions: Optional[Permissions]
    inheritPermissions: Optional[bool]
    isEncrypted: Optional[bool]
    encryptionInfo: Optional[EncryptionInfo]
    cntDeletedVersions: Optional[int]
    cntComments: Optional[int]
    cntDownloadShares: Optional[int]
    cntUploadShares: Optional[int]
    recycleBinRetentionPeriod: Optional[int]
    hasActivitiesLog: Optional[bool]
    quota: Optional[int]
    isFavorite: Optional[bool]
    branchVersion: Optional[int]
    mediaToken: Optional[str]
    isBrowsable: Optional[bool]
    cntRooms: Optional[int]
    cntFolders: Optional[int]
    cntFiles: Optional[int]
    authParentId: Optional[int]
    
class LogEvent(BaseModel):
    id: int
    time: datetime
    userId: int
    message: str
    operationId: Optional[int]
    operationName: Optional[str]
    status: Optional[int]
    userClient: Optional[str]
    customerId: Optional[int]
    userName: Optional[str]
    userIp: Optional[str]
    authParentSource: Optional[str]
    authParentTarget: Optional[str]
    objectId1: Optional[int]
    objectType1: Optional[int]
    objectName1: Optional[str]
    objectId2: Optional[int]
    objectType2: Optional[int]
    objectName2: Optional[str]
    attribute1: Optional[str]
    attribute2: Optional[str]
    attribute3: Optional[str]
    
class LogEventList(BaseModel):
    range: Range
    items: List[LogEvent]
    
class FileFileKeys(BaseModel):
    id: int
    fileKeyContainer: FileKey
    
class UserUserPublicKey(BaseModel):
    id: int
    publicKeyContainer: PublicKeyContainer
    
class UserIdFileIdItem(BaseModel):
    userId: int 
    fileId: int
    
class MissingKeysResponse(BaseModel):
    range: Range
    items: List[UserIdFileIdItem]
    users: List[UserUserPublicKey]
    files: List[FileFileKeys]
    
class FileVersion(BaseModel):
    id: int
    referenceId: int
    name: str
    parentId: Optional[int]
    deleted: Optional[bool]
    
class FileVersionList(BaseModel):
    range: Range
    items: List[FileVersion]
    

class Callback(Protocol):
    """ callback function signature - example see TransferJob.update_progress() """
    def __call__(self, val: int, total: int = ...) -> Any:
        ...
        
class TransferJob:
    """ object representing a single transfer (up- / download) """
    progress = 0
    transferred = 0
    total = 0
    
    def update_progress(self, val: int, total: int = None) -> None:
        self.transferred += val
        if total is not None and self.total == 0:
            # only set total if present and not set
            self.total = total
        
    @property
    def progress(self):
        if self.total > 0:
            return self.transferred / self.total
        else:
            return 0
































"""
Async DRACOON users adapter based on httpx and pydantic
V1.0.0
(c) Octavio Simone, November 2021 

Collection of DRACOON API calls for user management
Documentation: https://dracoon.team/api/swagger-ui/index.html?configUrl=/api/spec_v4/swagger-config#/users

Please note: maximum 500 items are returned in GET requests 

 - refer to documentation for details on filtering and offset 
 - use documentation for payload description 


"""

from typing import List
import logging

import httpx
from pydantic import validate_arguments

from dracoon.client import DRACOONClient, OAuth2ConnectionType
from dracoon.user.responses import (AttributesResponse, LastAdminUserRoomList, RoleList, 
                                    UserData, UserGroupList, UserList)
from dracoon.errors import ClientDisconnectedError, InvalidClientError
from .models import (AttributeEntry, CreateUser, Expiration, UpdateUser, UpdateUserAttributes, 
                     UserAuthData)


class DRACOONUsers:

    """
    API wrapper for DRACOON users endpoint:
    User management - requires user manager role.
    """

    def __init__(self, dracoon_client: DRACOONClient):
        """ requires a DRACOONClient to perform any request """

        if not isinstance(dracoon_client, DRACOONClient):
            raise InvalidClientError(message='Invalid client.')
        if dracoon_client.connection:
            self.dracoon = dracoon_client
            self.api_url = self.dracoon.base_url + self.dracoon.api_base_url + '/users'
            self.logger = logging.getLogger('dracoon.users')
            if self.dracoon.raise_on_err:
                self.raise_on_err = True
            else:
                self.raise_on_err = False
            self.logger.debug("DRACOON users adapter created.")
        else:
            self.logger.error("DRACOON client error: no connection. ")
            raise ClientDisconnectedError(message='DRACOON client must be connected: client.connect()')

    @validate_arguments
    async def create_user(self, user: CreateUser, raise_on_err: bool = False) -> UserData:
        """ creates a new user """

        payload = user.dict(exclude_unset=True)

        if not await self.dracoon.test_connection() and self.dracoon.connection:
            await self.dracoon.connect(OAuth2ConnectionType.refresh_token)

        try:
            res = await self.dracoon.http.post(self.api_url, json=payload)
            res.raise_for_status()
        except httpx.RequestError as e:
            await self.dracoon.handle_connection_error(e)
        except httpx.HTTPStatusError as e:
            self.logger.error("Creating user failed.")
            await self.dracoon.handle_http_error(err=e, raise_on_err=raise_on_err)
        
        self.logger.info("Created user.")
        return UserData(**res.json())

    def make_local_user(self, first_name: str, last_name: str, email: str, login: str = None,
                        language: str = None, notify: bool = None, expiration: Expiration = None, 
                        phone: str = None) -> CreateUser:
        """ makes a new local (basic) user required for create_user() """
        auth = self.make_auth_data(method='basic')

        user = {
            "firstName": first_name,
            "lastName": last_name, 
            "email": email,
            "userName": email,
            "authData": auth  
        }

        if login: user["userName"] = login
        if language: user["receiverLanguage"] = language
        if notify is not None: user["notifyUser"] = notify
        if expiration: user["expiration"] = expiration
        if phone: user["phone"] = phone

        return CreateUser(**user)

    def make_oidc_user(self, first_name: str, last_name: str, email: str, login: str, oidc_id: int, 
                       language: str = None, notify: bool = None, expiration: Expiration = None, 
                       phone: str = None, raise_on_err: bool = False) -> CreateUser:
        """ makes a new OpenID Connect (openid) user required for create_user() """
        auth = self.make_auth_data(method='openid', oidc_id=oidc_id, login=login)
           
        user = {
            "firstName": first_name,
            "lastName": last_name, 
            "email": email,
            "authData": auth  
        }    

        if language: user["receiverLanguage"] = language
        if notify is not None: user["notifyUser"] = notify
        if expiration: user["expiration"] = expiration
        if phone: user["phone"] = phone
        
        return CreateUser(**user)

    def make_ad_user(self, first_name: str, last_name: str, email: str, login: str, ad_id: int,
                     language: str = None, notify: bool = None, expiration: Expiration = None, 
                     phone: str = None) -> CreateUser:
        """ makes a new Active Directory (active_directory) user required for create_user() """
        auth = self.make_auth_data(method='active_directory', ad_id=ad_id, login=login)
        
        user = {
            "firstName": first_name,
            "lastName": last_name, 
            "email": email,
            "authData": auth,
            "userName": login 
        }    

        if language: user["receiverLanguage"] = language
        if notify is not None: user["notifyUser"] = notify
        if expiration: user["expiration"] = expiration
        if phone: user["phone"] = phone   
        
        return CreateUser(**user)

    def make_user_update(self, first_name: str = None, last_name: str = None, email: str = None, 
                         user_name: str = None, locked: bool = None, phone: str = None, 
                         expiration: Expiration = None, language: str = None,
                         auth_data: UserAuthData = None, 
                         non_member_viewer: bool = None) -> UpdateUser:
        """ makes an user update payload required for update_user() """
        user_update = {

            }
        
        if language: user_update["receiverLanguage"] = language
        if locked is not None: user_update["isLocked"] = locked
        if expiration: user_update["expiration"] = expiration
        if phone: user_update["phone"] = phone
        if first_name: user_update["firstName"] = first_name  
        if last_name: user_update["lastName"] = last_name   
        if non_member_viewer is not None: user_update["isNonmemberViewer"] = non_member_viewer   
        if auth_data: user_update["authData"] = auth_data   
        if email: user_update["email"] = email 
        if user_name: user_update["userName"] = user_name

        return UpdateUser(**user_update)

        
    def make_auth_data(self, method: str, login: str = None, password: str = None, 
                      change_password: bool = None, ad_id: int = None, 
                      oidc_id: int = None) -> UserAuthData:
        """ makes authentication data required for user creation / modification """
        auth =  {
            "method": method
        }

        if login: auth["login"] = login
        if password: auth["password"] = password
        if change_password is not None: auth["mustChangePassword"] = change_password
        if ad_id: auth["adConfigId"] = ad_id
        if oidc_id: auth["oidConfigId"] = oidc_id

        return UserAuthData(**auth)


    @validate_arguments
    async def get_users(self, offset: int = 0, filter: str = None, limit: int = None, 
                        sort: str = None, raise_on_err: bool = False) -> UserList:     
        """ list (all) users """
        if not await self.dracoon.test_connection() and self.dracoon.connection:
            await self.dracoon.connect(OAuth2ConnectionType.refresh_token)

        api_url = self.api_url + f'/?offset={offset}'
        if filter != None: api_url += f'&filter={filter}' 
        if limit != None: api_url += f'&limit={str(limit)}' 
        if sort != None: api_url += f'&sort={sort}' 

        try:
            res = await self.dracoon.http.get(api_url)
            res.raise_for_status()
        except httpx.RequestError as e:
            await self.dracoon.handle_connection_error(e)
        except httpx.HTTPStatusError as e:
            self.logger.error("Getting users failed.")
            await self.dracoon.handle_http_error(err=e, raise_on_err=raise_on_err)
        
        self.logger.info("Retrieved users.")
        return UserList(**res.json())

    @validate_arguments
    async def get_user(self, user_id: int, raise_on_err: bool = False) -> UserData:
        """ get user details for specific user (by id) """
        if not await self.dracoon.test_connection() and self.dracoon.connection:
            await self.dracoon.connect(OAuth2ConnectionType.refresh_token)

        api_url = self.api_url + f'/{str(user_id)}'

        try:
            res = await self.dracoon.http.get(api_url)
            res.raise_for_status()
        except httpx.RequestError as e:
            await self.dracoon.handle_connection_error(e)
        except httpx.HTTPStatusError as e:
            self.logger.error("Getting user failed.")
            await self.dracoon.handle_http_error(err=e, raise_on_err=raise_on_err)
        
        self.logger.info("Retrieved user.")
        return UserData(**res.json())

    # update user's meta data for given user id
    @validate_arguments
    async def update_user(self, user_id: int, user_update: UpdateUser, 
                          raise_on_err: bool = False) -> UserData:
        """ update user details for specific user (by id) """
        if not await self.dracoon.test_connection() and self.dracoon.connection:
            await self.dracoon.connect(OAuth2ConnectionType.refresh_token)

        api_url = self.api_url + f'/{str(user_id)}'

        payload = user_update.dict(exclude_unset=True)

        try:
            res = await self.dracoon.http.put(url=api_url, json=payload)
            res.raise_for_status()
        except httpx.RequestError as e:
            await self.dracoon.handle_connection_error(e)
        except httpx.HTTPStatusError as e:
            self.logger.error("Updating user failed.")
            await self.dracoon.handle_http_error(err=e, raise_on_err=raise_on_err)
        
        self.logger.info("Updated user.")
        return UserData(**res.json())

    # delete user for given user id
    @validate_arguments
    async def delete_user(self, user_id: int, raise_on_err: bool = False) -> None:
        """ delete specific user (by id) """
        if not await self.dracoon.test_connection() and self.dracoon.connection:
            await self.dracoon.connect(OAuth2ConnectionType.refresh_token)

        api_url = self.api_url + f'/{str(user_id)}'

        try:
            res = await self.dracoon.http.delete(api_url)
            res.raise_for_status()
        except httpx.RequestError as e:
            await self.dracoon.handle_connection_error(e)
        except httpx.HTTPStatusError as e:
            self.logger.error("Deleting user failed.")
            await self.dracoon.handle_http_error(err=e, raise_on_err=raise_on_err)
        
        self.logger.info("Deleted user.")
        return None

    # get user details for given user id
    @validate_arguments
    async def get_user_groups(self, user_id: int, offset: int = 0, filter: str = None, 
                              limit: int = None, sort: str = None, 
                              raise_on_err: bool = False) -> UserGroupList:
        """ list all groups for a specific user (by id) """
        if not await self.dracoon.test_connection() and self.dracoon.connection:
            await self.dracoon.connect(OAuth2ConnectionType.refresh_token)

        api_url = self.api_url + f'/{user_id}/groups/?offset={str(offset)}'
        if filter != None: api_url += f'&filter={filter}' 
        if limit != None: api_url += f'&limit={str(limit)}' 
        if sort != None: api_url += f'&sort={sort}' 

        try:
            res = await self.dracoon.http.get(api_url)
            res.raise_for_status()
        except httpx.RequestError as e:
            await self.dracoon.handle_connection_error(e)
        except httpx.HTTPStatusError as e:
            self.logger.error("Getting user groups failed.")
            await self.dracoon.handle_http_error(err=e, raise_on_err=raise_on_err)
        
        self.logger.info("Retrieved user groups.")
        return UserGroupList(**res.json())

    # get rooms in which user is last remaining admin (prevents user deletion!)
    @validate_arguments
    async def get_user_last_admin_rooms(self, user_id: int, 
                                        raise_on_err: bool = False) -> LastAdminUserRoomList:
        """ list all rooms, in which user is last admin (by id) """
        if not await self.dracoon.test_connection() and self.dracoon.connection:
            await self.dracoon.connect(OAuth2ConnectionType.refresh_token)

        api_url = self.api_url + f'/{str(user_id)}/last_admin_rooms'

        try:
            res = await self.dracoon.http.get(api_url)
            res.raise_for_status()
        except httpx.RequestError as e:
            await self.dracoon.handle_connection_error(e)
        except httpx.HTTPStatusError as e:
            self.logger.error("Getting user last admin rooms failed.")
            await self.dracoon.handle_http_error(err=e, raise_on_err=raise_on_err)

        self.logger.info("Retrieved user last admin rooms.")
        return LastAdminUserRoomList(**res.json())

    # get roles assigned to user
    @validate_arguments
    async def get_user_roles(self, user_id: int, raise_on_err: bool = False) -> RoleList:
        """ get user roles for specific user (by id) """
        if not await self.dracoon.test_connection() and self.dracoon.connection:
            await self.dracoon.connect(OAuth2ConnectionType.refresh_token)

        api_url = self.api_url + f'/{str(user_id)}/roles'

        try:
            res = await self.dracoon.http.get(api_url)
            res.raise_for_status()
        except httpx.RequestError as e:
            await self.dracoon.handle_connection_error(e)
        except httpx.HTTPStatusError as e:
            self.logger.error("Getting user roles failed.")
            await self.dracoon.handle_http_error(err=e, raise_on_err=raise_on_err)
        
        self.logger.info("Retrieved user roles.")
        return RoleList(**res.json())

    # get custom user attributes (key, value)
    @validate_arguments
    async def get_user_attributes(self, user_id: int, offset: int = 0, filter: str = None, 
                                  limit: int = None, sort: str = None, 
                                  raise_on_err: bool = False) -> AttributesResponse:
        """ get custom user attributes for a specific user (by id) """
        if not await self.dracoon.test_connection() and self.dracoon.connection:
            await self.dracoon.connect(OAuth2ConnectionType.refresh_token)

        api_url = self.api_url + f'/{user_id}/userAttributes/?offset={str(offset)}'
        if filter != None: api_url += f'&filter={filter}' 
        if limit != None: api_url += f'&limit={str(limit)}' 
        if sort != None: api_url += f'&sort={sort}' 

        try:
            res = await self.dracoon.http.get(api_url)
            res.raise_for_status()
        except httpx.RequestError as e:
            await self.dracoon.handle_connection_error(e)
        except httpx.HTTPStatusError as e:
            self.logger.error("Getting user attributes failed.")
            await self.dracoon.handle_http_error(err=e, raise_on_err=raise_on_err)
        
        self.logger.info("Retrieved user attributes.")
        return AttributesResponse(**res.json())

    # set custom user attributes (key, value)
    @validate_arguments
    async def delete_user_attribute(self, user_id: int, key: str, 
                                    raise_on_err: bool = False) -> None:
        """ delete custom user attribute for a specific user (by id) """
        if not await self.dracoon.test_connection() and self.dracoon.connection:
            await self.dracoon.connect(OAuth2ConnectionType.refresh_token)

        api_url = self.api_url + f'/{str(user_id)}/userAttributes/{key}'

        try:
            res = await self.dracoon.http.delete(api_url)
            res.raise_for_status()
        except httpx.RequestError as e:
            await self.dracoon.handle_connection_error(e)
        except httpx.HTTPStatusError as e:
            self.logger.error("Deleting user attribute failed.")
            await self.dracoon.handle_http_error(err=e, raise_on_err=raise_on_err)
        
        self.logger.info("Deleted user attribute.")
        return None

    # update custom user attributes (key, value)
    @validate_arguments
    async def update_user_attributes(self, user_id: int, attributes: UpdateUserAttributes, 
                                     raise_on_err: bool = False) -> UserData:
        """ create / update custom user attribute for a specific user (by id) """
        if not await self.dracoon.test_connection() and self.dracoon.connection:
            await self.dracoon.connect(OAuth2ConnectionType.refresh_token)

        api_url = self.api_url + f'/{str(user_id)}/userAttributes'
        payload = attributes.dict(exclude_unset=True)

        try:
            res = await self.dracoon.http.put(url=api_url, json=payload)
            res.raise_for_status()
        except httpx.RequestError as e:
            await self.dracoon.handle_connection_error(e)
        except httpx.HTTPStatusError as e:
            self.logger.error("Updating user attributes failed.")
            await self.dracoon.handle_http_error(err=e, raise_on_err=raise_on_err)

        self.logger.info("Updated user attributes.")
        return UserData(**res.json())

    def make_custom_user_attribute(self, key: str, value: str) -> AttributeEntry:
        """ make a custom user attribute required for update_user_attributes() """
        return AttributeEntry(**{
            "key": key,
            "value": value
        })

    def make_attributes_update(self, list: List[AttributeEntry]) -> UpdateUserAttributes:
        """ consolidate mulitple attributes into list - required for update_user_attributes() """
        return UpdateUserAttributes(**{
            "items": list
        })




"""
Async DRACOON eventlog adapter based on httpx and pydantic
V1.2.0
(c) Octavio Simone, November 2021 

Collection of DRACOON API calls for user management
Documentation: https://dracoon.team/api/swagger-ui/index.html?configUrl=/api/spec_v4/swagger-config#/eventlog

Please note: maximum 500 items are returned in GET requests 

 - refer to documentation for details on filtering and offset 
 - use documentation for payload description 


"""

from typing import List
import httpx
import logging
from pydantic import validate_arguments

from dracoon.client import DRACOONClient, OAuth2ConnectionType
from dracoon.errors import ClientDisconnectedError, InvalidClientError
from .responses import AuditNodeInfoResponse, AuditNodeResponse, LogEventList


class DRACOONEvents:

    """
    API wrapper for DRACOON eventlog endpoint:
    Events and permissions audit management - requires user auditor role.
    Please note: using the eventlog API for events is discouraged. Please 
    use the reports API wrapper instead.
    """

    def __init__(self, dracoon_client: DRACOONClient):
        """ requires a DRACOONClient to perform any request """
        if not isinstance(dracoon_client, DRACOONClient):
            raise InvalidClientError(message='Invalid client.')
        if dracoon_client.connection:

            self.dracoon = dracoon_client
            self.api_url = self.dracoon.base_url + self.dracoon.api_base_url + '/eventlog'
            self.logger = logging.getLogger('dracoon.eventlog')
            if self.dracoon.raise_on_err:
                self.raise_on_err = True
            else:
                self.raise_on_err = False

            self.logger.debug("DRACOON eventlog adapter created.")
        
        else:
            self.logger.critical("DRACOON client not connected.")
            raise ClientDisconnectedError(message='DRACOON client must be connected: client.connect()')
   
    @validate_arguments
    async def get_permissions(self, offset: int = 0, filter: str = None, limit: int = None, sort: str = None, raise_on_err = False) -> List[AuditNodeResponse]:
        """ get permissions for all nodes (rooms) """
        if not await self.dracoon.test_connection() and self.dracoon.connection:
            await self.dracoon.connect(OAuth2ConnectionType.refresh_token)

        if self.raise_on_err:
            raise_on_err = True

        api_url = self.api_url + f'/audits/nodes/?offset={offset}'
        if filter != None: api_url += f'&filter={filter}' 
        if limit != None: api_url += f'&limit={str(limit)}' 
        if sort != None: api_url += f'&sort={sort}' 

        try:
            res = await self.dracoon.http.get(api_url)
            res.raise_for_status()
        except httpx.RequestError as e:
            await self.dracoon.handle_connection_error(e)
        except httpx.HTTPStatusError as e:
            self.logger.error("Getting permissions failed.")
            await self.dracoon.handle_http_error(err=e, raise_on_err=raise_on_err)
        
        self.logger.info("Retrieved node permission audit.")
        tmp_list = res.json()
        
        return [AuditNodeResponse(**node_info) for node_info in tmp_list]
    
    @validate_arguments
    async def get_rooms(self, parent_id: int = 0, offset: int = 0, filter: str = None, 
                                   limit: int = None, sort: str = None, raise_on_err = False) -> AuditNodeInfoResponse:
        """ get permissions for all nodes (rooms) """
        if not await self.dracoon.test_connection() and self.dracoon.connection:
            await self.dracoon.connect(OAuth2ConnectionType.refresh_token)

        if self.raise_on_err:
            raise_on_err = True

        api_url = self.api_url + f'/audits/node_info/?parent_id={str(parent_id)}&offset={offset}'
        if filter != None: api_url += f'&filter={filter}' 
        if limit != None: api_url += f'&limit={str(limit)}' 
        if sort != None: api_url += f'&sort={sort}'

        try:
            res = await self.dracoon.http.get(api_url)
            res.raise_for_status()
        except httpx.RequestError as e:
            await self.dracoon.handle_connection_error(e)
        except httpx.HTTPStatusError as e:
            self.logger.error("Getting permissions failed.")
            await self.dracoon.handle_http_error(err=e, raise_on_err=raise_on_err)
        
        self.logger.info("Retrieved node permission audit.")
        return AuditNodeInfoResponse(**res.json())

    @validate_arguments
    async def get_events(self, offset: int = 0, filter: str = None, limit: int = None, 
                        sort: str = None, date_start: str = None, date_end: str = None, operation_id: int = None, user_id: int = None, raise_on_err = False) -> LogEventList:
        """ get events (audit log) """
        if not await self.dracoon.test_connection() and self.dracoon.connection:
            await self.dracoon.connect(OAuth2ConnectionType.refresh_token)
        
        if self.raise_on_err:
            raise_on_err = True

        api_url = self.api_url + f'/events/?offset={offset}'
        if date_start != None: api_url += f'&date_start={date_start}'
        if date_end != None: api_url += f'&date_end={date_end}'
        if operation_id != None: api_url += f'&type={str(operation_id)}'
        if user_id != None: api_url += f'&user_id={str(user_id)}'
        if filter != None: api_url += f'&filter={filter}' 
        if limit != None: api_url += f'&limit={str(limit)}' 
        if sort != None: api_url += f'&sort={sort}' 

        try:
            res = await self.dracoon.http.get(api_url)
            res.raise_for_status()
        except httpx.RequestError as e:
            await self.dracoon.handle_connection_error(e)

        except httpx.HTTPStatusError as e:
            self.logger.error("Getting event log failed.")
            await self.dracoon.handle_http_error(err=e, raise_on_err=raise_on_err)

        self.logger.info("Retrieved events from eventlog.")
        return LogEventList(**res.json())





from dracoon.client import DRACOONClient, DRACOONConnection, OAuth2ConnectionType
import unittest
import dotenv
import os


dotenv.load_dotenv()

client_id = os.environ.get('CLIENT_ID')
client_secret = os.environ.get('CLIENT_SECRET')
username = os.environ.get('E2E_USER_NAME')
password = os.environ.get('E2E_PASSWORD')
base_url = os.environ.get('E2E_BASE_URL')

class TestDRACOONClient(unittest.TestCase):

    def test_client_creation_default_oauth(self):

        dracoon = DRACOONClient(base_url='https://just.a.test.com')

        self.assertEqual(dracoon.client_id, 'dracoon_legacy_scripting')
        self.assertEqual(dracoon.client_secret, '')
        self.assertFalse(dracoon.connected)


    def test_client_creation_custom_oauth(self):

        dracoon = DRACOONClient(base_url='https://just.a.test.com', client_id='my_test_client', client_secret='my_test_secret')

        self.assertEqual(dracoon.client_id, 'my_test_client')
        self.assertEqual(dracoon.client_secret, 'my_test_secret')
        self.assertFalse(dracoon.connected)

    
class TestAsyncDRACOONClient(unittest.IsolatedAsyncioTestCase):
    async def test_connection_offline(self):

        dracoon = DRACOONClient(base_url='https://just.a.test.com', client_id='my_test_client', client_secret='my_test_secret')

        self.assertFalse(await dracoon.test_connection())

    
    async def test_connection_password_flow(self):

        dracoon = DRACOONClient(base_url=base_url, client_id=client_id, client_secret=client_secret)

        await dracoon.connect(OAuth2ConnectionType.password_flow, username=username, password=password)

        self.assertTrue(await dracoon.test_connection())
        self.assertIsNotNone(dracoon.connection)
        self.assertIsInstance(dracoon.connection, DRACOONConnection)
        self.assertTrue(dracoon.connected)

        await dracoon.logout(revoke_refresh_token=True)

    
    async def test_refresh_token_auth(self):

        dracoon = DRACOONClient(base_url=base_url, client_id=client_id, client_secret=client_secret)

        connection1 = await dracoon.connect(OAuth2ConnectionType.password_flow, username=username, password=password)
        connection2 = await dracoon.connect(OAuth2ConnectionType.refresh_token)

        self.assertIsInstance(connection1, DRACOONConnection)
        self.assertIsInstance(connection2, DRACOONConnection)

        self.assertNotEqual(connection1, connection2)
        self.assertTrue(await dracoon.test_connection())
        self.assertTrue(dracoon.connected)

        await dracoon.logout(revoke_refresh_token=True)

    async def test_logout(self):
        dracoon = DRACOONClient(base_url=base_url, client_id=client_id, client_secret=client_secret)

        await dracoon.connect(OAuth2ConnectionType.password_flow, username=username, password=password)

        await dracoon.logout()

        self.assertFalse(dracoon.connected)
        self.assertIsNone(dracoon.connection)

    
    async def test_access_token_validity(self):
        dracoon = DRACOONClient(base_url=base_url, client_id=client_id, client_secret=client_secret)

        await dracoon.connect(OAuth2ConnectionType.password_flow, username=username, password=password)

        access_token_valid_no_test = await dracoon.check_access_token()
        access_token_valid_test = await dracoon.check_access_token(test=True)

        self.assertTrue(access_token_valid_no_test)
        self.assertTrue(access_token_valid_test)

        await dracoon.logout()
if __name__ == '__main__':
    unittest.main()


from dataclasses import dataclass
from pydantic import BaseModel
from typing import Optional
from datetime import datetime
from enum import Enum

### Enums 

# File key version (2048 or 4096 bit)
class FileKeyVersion(Enum):
    RSA2048_AES256GCM = 'A'
    RSA_4096_AES256GCM = 'RSA-4096/AES-256-GCM'

# Plain file key version (AES256GCM)
class PlainFileKeyVersion(Enum):
    AES256GCM = 'AES-256-GCM'

# User keypair (2048 or 4096 bit)
class UserKeyPairVersion(Enum):
    RSA2048 = 'A'
    RSA4096 = 'RSA-4096'


### Models

# file key AES256
class FileKey(BaseModel):
    key: str
    iv: str
    version: FileKeyVersion
    tag: Optional[str]

    class Config:
        use_enum_values = True

class PlainFileKey(BaseModel):
    version: PlainFileKeyVersion
    key: str
    iv: str
    tag: Optional[str]

    class Config:
        use_enum_values = True

class PublicKeyContainer(BaseModel):
    version: UserKeyPairVersion
    publicKey: str
    createdAt: Optional[datetime]
    expireAt: Optional[datetime]
    createdBy: Optional[int]
    class Config:
        use_enum_values = True

class PrivateKeyContainer(BaseModel):
    version: UserKeyPairVersion
    privateKey: str
    createdAt: Optional[datetime]
    expireAt: Optional[datetime]
    createdBy: Optional[int]

    class Config:
        use_enum_values = True

class UserKeyPairContainer(BaseModel):
    privateKeyContainer: PrivateKeyContainer
    publicKeyContainer: PublicKeyContainer

class PlainUserKeyPairContainer(BaseModel):
    privateKeyContainer: PrivateKeyContainer
    publicKeyContainer: PublicKeyContainer

class KeyState(Enum):
    none = "none"
    available = "available"
    pending = "pending"


@dataclass
class EncryptionInfo:
    userKeyState: KeyState
    roomKeyState: KeyState
    dataSpaceKeyState: KeyState
    
    class Config:
        use_enum_values = True






from dracoon.groups import DRACOONGroups
from dracoon import DRACOONClient, OAuth2ConnectionType
import dotenv
import os
import asyncio
from dracoon.groups.models import CreateGroup, UpdateGroup

from dracoon.groups.responses import Group, GroupList, GroupUserList, LastAdminGroupRoomList
from dracoon.user.responses import RoleList

dotenv.load_dotenv()

client_id = os.environ.get('CLIENT_ID')
client_secret = os.environ.get('CLIENT_SECRET')
username = os.environ.get('E2E_USER_NAME')
password = os.environ.get('E2E_PASSWORD')
base_url = os.environ.get('E2E_BASE_URL')
USER_ID = 795 # replace with existing user id

async def test_groups_e2e():

    dracoon = DRACOONClient(base_url=base_url, client_id=client_id, client_secret=client_secret)
    await dracoon.connect(OAuth2ConnectionType.password_flow, username=username, password=password)
    groups = DRACOONGroups(dracoon_client=dracoon)
    assert isinstance(groups, DRACOONGroups)
    print('Connection test complete (/)')


    group_list = await groups.get_groups()
    assert isinstance(group_list, GroupList)
    print('Getting user list test passed (/)')

    new_group = groups.make_group(name='THIS IS A TEST')
    assert isinstance(new_group, CreateGroup)

    print('Making group create payload test passed (/)')

    group = await groups.create_group(new_group)
    assert isinstance(group, Group)

    print('Creating a group test passed (/)')

    update = groups.make_group_update(name='THIS IS ANOTHER TEST')
    assert isinstance(update, UpdateGroup)

    print('Making group update payload test passed (/)')

    updated_group = await groups.update_group(group_id=group.id, group_update=update)
    assert updated_group.name == 'THIS IS ANOTHER TEST'
    print('Updating a group test passed (/)')

    group_roles = await groups.get_group_roles(group_id=group.id)
    assert isinstance(group_roles, RoleList)
    print('Getting group roles test passed (/)')

    group_last_admin_rooms = await groups.get_group_last_admin_rooms(group_id=group.id)
    assert isinstance(group_last_admin_rooms, LastAdminGroupRoomList)
    print('Getting group last admin rooms test passed (/)')
    
    user_list = [USER_ID]
    add_user = await groups.add_group_users(group_id=group.id, user_list=user_list)
    assert isinstance(add_user, Group)
    print('Adding user to group test passed (/)')

    group_users = await groups.get_group_users(group_id=group.id)
    assert isinstance(group_users, GroupUserList)
    assert len(group_users.items) > 0
    assert group_users.items[0].userInfo.id == USER_ID
    print('Getting group users test passed (/)')

    del_group_user = await groups.delete_group_users(group_id=group.id, user_list=user_list)
    assert isinstance(del_group_user, Group)
    print('Deleting group users test passed (/)')

    del_group = await groups.delete_group(group_id=group.id)

    assert del_group == None

    print('Deleting group test passed (/)')

    print('Groups E2E tests passed (/)')



if __name__ == '__main__':
    asyncio.run(test_groups_e2e())

    

   





"""
Async DRACOON API wrapper based on httpx and pydantic
V1.2.0
(c) Octavio Simone, November 2021 

Collection of DRACOON API calls
Documentation: https://dracoon.team/api

Please note: maximum 500 items are returned in GET requests 

 - refer to documentation for details on filtering and offset 
 - use documentation for payload description 
 
All requests with bodies use generic params variable to pass JSON body

"""


import logging
import asyncio
from pathlib import Path
from typing import Any, Generator, List
from datetime import datetime
from dracoon.client.models import ProxyConfig
from dracoon.nodes.models import Callback


from dracoon.nodes.responses import S3FileUploadStatus

from .crypto.models import PlainUserKeyPairContainer
from .downloads import DRACOONDownloads
from .public import DRACOONPublic
from .client import DRACOONClient, DRACOONConnection, OAuth2ConnectionType
from .eventlog import DRACOONEvents
from .nodes import CHUNK_SIZE, MIN_CHUNK_SIZE, DRACOONNodes
from .shares import DRACOONShares
from .user import DRACOONUser
from .users import DRACOONUsers
from .groups import DRACOONGroups
from .settings import DRACOONSettings
from .reports import DRACOONReports
from .crypto import decrypt_private_key
from .logger import create_logger
from .errors import (CryptoMissingFileKeyrError, CryptoMissingKeypairError,
                     HTTPNotFoundError, InvalidArgumentError, InvalidFileError, InvalidPathError, ClientDisconnectedError)



class DRACOON:
    """ DRACOON main API wrapper with all adapters to specific endpoints """ 

    def __init__(self, base_url: str, client_id: str = 'dracoon_legacy_scripting', client_secret: str = '', log_file: str = 'dracoon.log', 
                 log_level = logging.INFO, log_stream: bool = False, raise_on_err: bool = False, proxy_config: ProxyConfig = None,
                 log_file_out: bool = False):
        """ intialize with instance information: base DRACOON url and OAuth app client credentials """
        self.client = DRACOONClient(base_url=base_url, client_id=client_id, client_secret=client_secret, raise_on_err=raise_on_err, 
                                    proxy_config=proxy_config)
        self.logger = create_logger(log_file=log_file, log_level=log_level, log_stream=log_stream, log_file_out=log_file_out)
        self.logger.info("Created DRACOON client.")
        self.plain_keypair = None
        self.user_info =  None
  
    @property 
    def nodes(self) -> DRACOONNodes:
        return DRACOONNodes(self.client)
    
    @property  
    def public(self) -> DRACOONPublic:
        return DRACOONPublic(self.client)
    
    @property 
    def user(self) -> DRACOONUser:
        return DRACOONUser(self.client)
    
    @property 
    def reports(self) -> DRACOONReports:
        return DRACOONReports(self.client)
    
    @property 
    def settings(self) -> DRACOONSettings:
        return DRACOONSettings(self.client)
    
    @property
    def users(self) -> DRACOONUsers:
       return DRACOONUsers(self.client) 
   
    @property 
    def groups(self) -> DRACOONGroups:
        return DRACOONGroups(self.client)
    
    @property 
    def eventlog(self) -> DRACOONEvents:
        return DRACOONEvents(self.client)
    
    @property
    def shares(self) -> DRACOONShares:
       return DRACOONShares(self.client) 
   
    @property
    def downloads(self) -> DRACOONDownloads:
       return DRACOONDownloads(self.client) 
   
   
    async def connect(self, connection_type: OAuth2ConnectionType = OAuth2ConnectionType.auth_code, username: str = None, 
                      password: str = None, auth_code: str = None, refresh_token: str = None, redirect_uri: str = None) -> DRACOONConnection:
        """ establishes a connection required for all adapters """
        self.connection = await self.client.connect(connection_type=connection_type, username=username, password=password, 
                                               auth_code=auth_code, refresh_token=refresh_token, redirect_uri=redirect_uri)

        self.logger.info("Initialized DRACOON adapters.") 

        user_info_res = self.user.get_account_information()
        system_info_res = self.public.get_system_info()
        oidc_auth_info_res = self.public.get_auth_openid_info()
        ad_auth_info_res = self.public.get_auth_ad_info()

        self.logger.debug("Getting DRACOON instance information...")

        list = await asyncio.gather(user_info_res, system_info_res, oidc_auth_info_res, ad_auth_info_res)
        
        self.user_info = list[0]
        self.system_info = list[1]
        self.auth_ad_info = list[3]
        self.auth_oidc_info = list[2]

        self.logger.info("Retrieved instance and account information.")
        self.logger.debug("Logged in as user id %s.", self.user_info.id)
        self.logger.debug("Using S3: %s.", self.system_info.useS3Storage)
        
        return self.connection
        
    async def logout(self, revoke_refresh_token: bool = False) -> None:
        """ closes the httpx client and revokes tokens """
        await self.client.logout(revoke_refresh_token=revoke_refresh_token)
        self.logger.info("Revoked token(s).")
        self.logger.debug("Refresh token revoked: %s", revoke_refresh_token)

    async def test_connection(self) -> bool:
        """ test authenticated connection via authenticated ping """
        self.logger.debug("Testing authenticated connection.")
        return await self.client.test_connection()
    
    def valid_access_token(self) -> bool:
        """ check access token validity based on expiration """
        self.logger.debug("Checking access token validity.")
        return self.client.check_access_token()
    
    def valid_refresh_token(self) -> bool:
        """ check refresh token validity based on expiration """
        self.logger.debug("Checking refresh token validity.")
        return self.client.check_refresh_token()

    def check_keypair(self) -> bool:
        
        if not self.plain_keypair:
            return False
        
        self.logger.info("Checking user keypair.")
        return self.plain_keypair is not None and self.user_info is not None

    async def get_keypair(self, secret: str) -> PlainUserKeyPairContainer:
        """ get user keypair """
        if not self.client.connection:
            self.logger.error("DRACOON client not connected: Keypair not retrieved.")
            raise ClientDisconnectedError()

        enc_keypair = await self.user.get_user_keypair()
      
        plain_keypair = decrypt_private_key(secret, enc_keypair)

        self.plain_keypair = plain_keypair

        self.logger.info("Retrieved user keypair.")
        self.logger.debug("User keypair version: %s", self.plain_keypair.publicKeyContainer.version)

        return plain_keypair

    async def upload(self, file_path: str, target_path: str = None, resolution_strategy: str = 'autorename', 
                     display_progress: bool = False, modification_date: str = None, 
                     creation_date: str = None, 
                     raise_on_err: bool = False, callback_fn: Callback  = None,
                     target_parent_id: int = None,
                     chunksize: int = CHUNK_SIZE
                     ) -> S3FileUploadStatus:  
        """ upload a file to a target """
        if not self.client.connection:
            self.logger.error("DRACOON client not connected: Upload failed.")
            err = ClientDisconnectedError(message="DRACOON client not connected.")
            await self.client.handle_generic_error(err=err)
            
        if self.client.raise_on_err:
            raise_on_err = True
            
        if target_parent_id is None and target_path is None:
            self.logger.critical('Upload failed: Missing target info - id or path must be provided.')
            err = InvalidPathError(message='Missing mandatory arguments: target path or target parent id')
            await self.client.handle_generic_error(err=err)
            
            
        if target_parent_id is not None:
            node_info = await self.nodes.get_node(node_id=target_parent_id, raise_on_err=raise_on_err)
        elif target_path is not None:
            node_info = await self.nodes.get_node_from_path(path=target_path, raise_on_err=raise_on_err)
        else:
            raise InvalidArgumentError("Missing node info: Provide target path or node id.")
        
        if not node_info:
            self.logger.critical('Upload failed: Invalid target path.')
            msg = 'Node %s not found.', target_path
            self.logger.debug(msg)
            err = InvalidPathError(message=msg)
            await self.client.handle_generic_error(err=err)
        
        file_name = Path(file_path).name
        file_modified = Path(file_path).stat().st_mtime
        file_created = Path(file_path).stat().st_ctime
        
        if modification_date is None:
            modification_date = datetime.fromtimestamp(file_modified).strftime('%Y-%m-%dT%H:%M:%S')
            
        if creation_date is None:
            creation_date = datetime.fromtimestamp(file_created).strftime('%Y-%m-%dT%H:%M:%S')

           
        self.logger.info("Uploading file.")
        self.logger.debug("File: %s", file_path)
        self.logger.debug("Destination: %s", target_path)

        target_id = node_info.id
        is_encrypted = node_info.isEncrypted

        self.logger.debug("Encrypted: %s", is_encrypted)
        
        use_s3_storage = False
        
        if self.system_info.useS3Storage:
            use_s3_storage = True    
            if chunksize < MIN_CHUNK_SIZE: chunksize = MIN_CHUNK_SIZE

        self.logger.debug("Using S3 storage: %s", use_s3_storage)
            
        upload_channel_payload = self.nodes.make_upload_channel(parent_id=target_id, name=file_name, direct_s3_upload=use_s3_storage, 
                                                                modification_date=modification_date, creation_date=creation_date)
        upload_channel = await self.nodes.create_upload_channel(upload_channel=upload_channel_payload, raise_on_err=raise_on_err)
    

        self.logger.debug("Created upload channel: %s", upload_channel.uploadId)
        self.logger.debug("Upload URL (redacted): %s", upload_channel.uploadUrl[:-4])

        # crypto upload 
        if is_encrypted and self.check_keypair() and not use_s3_storage:       
            upload = await self.nodes.upload_encrypted(file_path=file_path, upload_channel=upload_channel, display_progress=display_progress,
                                                         plain_keypair=self.plain_keypair, resolution_strategy=resolution_strategy, 
                                                         raise_on_err=raise_on_err, callback_fn=callback_fn, chunksize=chunksize)
        elif is_encrypted and self.check_keypair() and use_s3_storage:
            upload = await self.nodes.upload_s3_encrypted(file_path=file_path, upload_channel=upload_channel, plain_keypair=self.plain_keypair, 
                                                          display_progress=display_progress, resolution_strategy=resolution_strategy, 
                                                          raise_on_err=raise_on_err, callback_fn=callback_fn, chunksize=chunksize)
        elif is_encrypted and not self.check_keypair():
            self.logger.critical("Upload failed: Keypair not unlocked.")
            raise CryptoMissingKeypairError('DRACOON crypto upload requires unlocked keypair. Please unlock keypair first.')
        # unencrypted upload
        elif not is_encrypted and not use_s3_storage:
            upload = await self.nodes.upload_unencrypted(file_path=file_path, upload_channel=upload_channel, 
                                                         resolution_strategy=resolution_strategy, display_progress=display_progress, 
                                                         raise_on_err=raise_on_err, callback_fn=callback_fn, chunksize=chunksize)
        elif not is_encrypted and use_s3_storage:
            upload = await self.nodes.upload_s3_unencrypted(file_path=file_path, upload_channel=upload_channel, 
                                                            display_progress=display_progress, resolution_strategy=resolution_strategy,
                                                            raise_on_err=raise_on_err, callback_fn=callback_fn, chunksize=chunksize)

        self.logger.info("Upload completed.")
        
        return upload

    async def download(self, target_path: str, file_path: str = None, display_progress: bool = False, raise_on_err: bool = False, 
                       callback_fn: Callback  = None, file_name: str = None, source_node_id: int = None, chunksize: int = CHUNK_SIZE):
        """ download a file to a target """

        if not self.client.connection:
            await self.logout()
            self.logger.error("DRACOON client not connected: Download failed.")
            raise ClientDisconnectedError(message='DRACOON client not connected.')

        if self.client.raise_on_err:
            raise_on_err = True
            
        if source_node_id is not None:
            node_id = source_node_id
            node_info = await self.nodes.get_node(node_id=node_id)   
        elif file_path is not None:
            node_info = await self.nodes.get_node_from_path(path=file_path, raise_on_err=raise_on_err)
        else:
            raise InvalidArgumentError("Missing node info: provide either id or file path")
        
        if not node_info:
            self.logger.error("Download failed: file does not exist.")
            self.logger.debug("File %s not found", file_path)
            err = InvalidFileError(message='File does not exist.')
            await self.client.handle_generic_error(err=err)
        
        node_id = node_info.id

        is_encrypted = node_info.isEncrypted

        self.logger.info("Downloading file.")
        self.logger.debug("File: %s", file_path)
        self.logger.debug("Destination: %s", target_path)
        self.logger.debug("Encrypted: %s", is_encrypted)

        dl_token_res = await self.nodes.get_download_url(node_id=node_id, raise_on_err=raise_on_err)

        download_url = dl_token_res.downloadUrl

        self.logger.debug("Download URL (redacted): %s", download_url[:-4])

        if not is_encrypted:
            await self.downloads.download_unencrypted(download_url=download_url, target_path=target_path, node_info=node_info, 
                                                      display_progress=display_progress, raise_on_err=raise_on_err, 
                                                      callback_fn=callback_fn, file_name=file_name, chunksize=chunksize)
        elif is_encrypted and self.check_keypair():
            try:
                file_key = await self.nodes.get_user_file_key(node_id, raise_on_err=True)
                await self.downloads.download_encrypted(download_url=download_url, target_path=target_path, node_info=node_info, 
                                                    plain_keypair=self.plain_keypair, file_key=file_key, display_progress=display_progress, 
                                                    raise_on_err=raise_on_err, callback_fn=callback_fn, file_name=file_name, chunksize=chunksize)
            except HTTPNotFoundError:
                raise CryptoMissingFileKeyrError(message=f'No file key for node {node_id}')
                
        elif is_encrypted and not self.check_keypair():
            raise CryptoMissingKeypairError(message='Keypair must be entered for encrypted nodes.')


    def get_code_url(self) -> str:
        """ get code url for authorization code flow """
        self.logger.info("Getting authorization URL.")
        return self.client.get_code_url()

    def batch_process(self, coro_list: List[Any], batch_size: int = 5) -> Generator[Any, None, None]:
        """ 
        helper method which returns a generator for a list 
        of couroutines 
        utility to process multiple requests async
        """ 
        return (coro_list[i:i + batch_size]  for i in range(0, len(coro_list), batch_size))



import os
import io
from setuptools import setup, find_packages

# The directory containing this file
DESCRIPTION = "DRACOON API wrapper in Python"
here = os.path.abspath(os.path.dirname(__file__))

# The text of the README file
try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        README = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

# This call to setup() does all the work
setup(
    name="dracoon",
    version="1.5.1",
    description=DESCRIPTION,
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/unbekanntes-pferd/DRACOON-PYTHON-API",
    author="Octavio Simone",
    author_email="octsim@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
    ],
    packages=find_packages(),
    include_package_data=True,
    install_requires=["httpx", "asyncio", "pydantic", "cryptography", "tqdm"]
)


"""
Async DRACOON groups adapter based on httpx and pydantic
V1.2.0
(c) Octavio Simone, November 2021 

Collection of DRACOON API calls for user management
Documentation: https://dracoon.team/api/swagger-ui/index.html?configUrl=/api/spec_v4/swagger-config#/groups

Please note: maximum 500 items are returned in GET requests 

 - refer to documentation for details on filtering and offset 
 - use documentation for payload description 
 

"""

import logging
from typing import List

import httpx
from pydantic import validate_arguments

from dracoon.user.responses import RoleList
from dracoon.client import DRACOONClient, OAuth2ConnectionType
from dracoon.errors import ClientDisconnectedError, InvalidClientError
from .models import CreateGroup, Expiration, UpdateGroup
from .responses import Group, GroupList, GroupUserList, LastAdminGroupRoomList


class DRACOONGroups:

    """
    API wrapper for DRACOON groups endpoint:
    Group management - requires group manager role.
    """

    def __init__(self, dracoon_client: DRACOONClient):
        """ requires a DRACOONClient to perform any request """
        if not isinstance(dracoon_client, DRACOONClient):
            raise InvalidClientError(message='Invalid client.')
        if dracoon_client.connection:

            self.dracoon = dracoon_client
            self.api_url = self.dracoon.base_url + self.dracoon.api_base_url + '/groups'
            self.logger = logging.getLogger('dracoon.groups')
            if self.dracoon.raise_on_err:
                self.raise_on_err = True
            else:
                self.raise_on_err = False

            self.logger.debug("DRACOON groups adapter created.")
        
        else:
            self.logger.error("DRACOON client error: no connection. ")
            raise ClientDisconnectedError(message='DRACOON client must be connected: client.connect()')

    @validate_arguments
    async def create_group(self, group: CreateGroup, raise_on_err: bool = False) -> Group:
        """ creates a new group """

        if self.raise_on_err:
            raise_on_err = True

        payload = group.dict(exclude_unset=True)

        if not await self.dracoon.test_connection() and self.dracoon.connection:
            await self.dracoon.connect(OAuth2ConnectionType.refresh_token)

        try:
            res = await self.dracoon.http.post(self.api_url, json=payload)
            res.raise_for_status()
        except httpx.RequestError as e:
            await self.dracoon.handle_connection_error(e)
        except httpx.HTTPStatusError as e:
            self.logger.error("Creating group failed.")
            await self.dracoon.handle_http_error(err=e, raise_on_err=raise_on_err)
        
        self.logger.info("Group created.")
        return Group(**res.json())

    def make_group(self, name: str, expiration: Expiration = None) -> CreateGroup:
        """ makes a group required for create_group() """

        group = {
            "name": name
        }
        
        if expiration: group["expiration"] = group

        return CreateGroup(**group)
    
    @validate_arguments
    async def get_groups(self, offset: int = 0, filter: str = None, limit: int = None, sort: str = None, raise_on_err: bool = False) -> GroupList:
        """ list (all) groups """
        if not await self.dracoon.test_connection() and self.dracoon.connection:
            await self.dracoon.connect(OAuth2ConnectionType.refresh_token)
        
        if self.raise_on_err:
            raise_on_err = True

        api_url = self.api_url + f'/?offset={offset}'
        if filter != None: api_url += f'&filter={filter}' 
        if limit != None: api_url += f'&limit={str(limit)}' 
        if sort != None: api_url += f'&sort={sort}' 

        try:
            res = await self.dracoon.http.get(api_url)
            res.raise_for_status()
        except httpx.RequestError as e:
            await self.dracoon.handle_connection_error(e)
        except httpx.HTTPStatusError as e:
            self.logger.error("Getting groups failed.")
            await self.dracoon.handle_http_error(err=e, raise_on_err=raise_on_err)
        
        self.logger.info("Retrieved groups.")
        return GroupList(**res.json())


    @validate_arguments
    async def get_group(self, group_id: int, raise_on_err: bool = True) -> Group:
        """ get user details for specific group (by id) """
        if not await self.dracoon.test_connection() and self.dracoon.connection:
            await self.dracoon.connect(OAuth2ConnectionType.refresh_token)

        if self.raise_on_err:
            raise_on_err = True

        api_url = self.api_url + f'/{str(group_id)}'

        try:
            res = await self.dracoon.http.get(api_url)
            res.raise_for_status()
        except httpx.RequestError as e:
            await self.dracoon.handle_connection_error(e)
        except httpx.HTTPStatusError as e:
            self.logger.error("Getting group failed.")
            await self.dracoon.handle_http_error(err=e, raise_on_err=raise_on_err)
        
        self.logger.info("Retrieved group.")
        return Group(**res.json())


    @validate_arguments
    async def update_group(self, group_id: int, group_update: UpdateGroup, raise_on_err: bool = False) -> Group:
        """ update user details for specific group (by id) """
        if not await self.dracoon.test_connection() and self.dracoon.connection:
            await self.dracoon.connect(OAuth2ConnectionType.refresh_token)

        if self.raise_on_err:
            raise_on_err = True

        api_url = self.api_url + f'/{str(group_id)}'

        payload = group_update.dict(exclude_unset=True)

        try:
            res = await self.dracoon.http.put(url=api_url, json=payload)
            res.raise_for_status()
        except httpx.RequestError as e:
            await self.dracoon.handle_connection_error(e)
        except httpx.HTTPStatusError as e:
            self.logger.error("Updating group failed.")
            await self.dracoon.handle_http_error(err=e, raise_on_err=raise_on_err)
        
        self.logger.info("Updated group.")
        return Group(**res.json())

    def make_group_update(self, name: str = None, expiration: Expiration = None) -> UpdateGroup:
        """ make a group update payload required for update_group() """
        group_update = {}

        if name: group_update["name"] = name
        if expiration: group_update["expiration"] = name

        return UpdateGroup(**group_update)

    @validate_arguments
    async def delete_group(self, group_id: int, raise_on_err = False) -> None:
        """ delete specific user (by id) """
        if not await self.dracoon.test_connection() and self.dracoon.connection:
            await self.dracoon.connect(OAuth2ConnectionType.refresh_token)

        if self.raise_on_err:
            raise_on_err = True

        api_url = self.api_url + f'/{str(group_id)}'

        try:
            res = await self.dracoon.http.delete(api_url)
            res.raise_for_status()
        except httpx.RequestError as e:
            await self.dracoon.handle_connection_error(e)
        except httpx.HTTPStatusError as e:
            self.logger.error("Deleting group failed.")
            await self.dracoon.handle_http_error(err=e, raise_on_err=raise_on_err)
        
        self.logger.info("Deleted group.")
        return None

    # get user details for given user id
    @validate_arguments
    async def get_group_users(self, group_id: int, offset: int = 0, filter: str = None, limit: int = None, sort: str = None, raise_on_err: bool = False) -> GroupUserList:
        """ list all users for a specific group (by id) """
        if not await self.dracoon.test_connection() and self.dracoon.connection:
            await self.dracoon.connect(OAuth2ConnectionType.refresh_token)

        if self.raise_on_err:
            raise_on_err = True

        api_url = self.api_url + f'/{group_id}/users/?offset={str(offset)}'
        if filter != None: api_url += f'&filter={filter}' 
        if limit != None: api_url += f'&limit={str(limit)}' 
        if sort != None: api_url += f'&sort={sort}' 

        try:
            res = await self.dracoon.http.get(api_url)
            res.raise_for_status()
        except httpx.RequestError as e:
            await self.dracoon.handle_connection_error(e)
        except httpx.HTTPStatusError as e:
            self.logger.error("Getting group users failed.")
            await self.dracoon.handle_http_error(err=e, raise_on_err=raise_on_err)
        
        self.logger.info("Retrieved group users.")
        return GroupUserList(**res.json())

    # get rooms in which group is last remaining admin (prevents user deletion!)
    @validate_arguments
    async def get_group_last_admin_rooms(self, group_id: int, raise_on_err: bool = False) -> LastAdminGroupRoomList:
        """ list all rooms, in which group is last admin (by id) """
        if not await self.dracoon.test_connection() and self.dracoon.connection:
            await self.dracoon.connect(OAuth2ConnectionType.refresh_token)

        if self.raise_on_err:
            raise_on_err = True

        api_url = self.api_url + f'/{str(group_id)}/last_admin_rooms'

        try:
            res = await self.dracoon.http.get(api_url)
            res.raise_for_status()
        except httpx.RequestError as e:
            await self.dracoon.handle_connection_error(e)
        except httpx.HTTPStatusError as e:
            self.logger.error("Getting group last admin rooms failed.")
            await self.dracoon.handle_http_error(err=e, raise_on_err=raise_on_err)
        
        self.logger.info("Retrieved last admin group list.")
        return LastAdminGroupRoomList(**res.json())

    # get roles assigned to group
    @validate_arguments
    async def get_group_roles(self, group_id: int, raise_on_err: bool = False) -> RoleList:
        """ get group roles for specific user (by id) """
        if not await self.dracoon.test_connection() and self.dracoon.connection:
            await self.dracoon.connect(OAuth2ConnectionType.refresh_token)

        if self.raise_on_err:
            raise_on_err = True

        api_url = self.api_url + f'/{str(group_id)}/roles'

        try:
            res = await self.dracoon.http.get(api_url)

            res.raise_for_status()
        except httpx.RequestError as e:
            await self.dracoon.handle_connection_error(e)
        except httpx.HTTPStatusError as e:
            self.logger.error("Getting group roles failed.")
            await self.dracoon.handle_http_error(err=e, raise_on_err=raise_on_err)
        
        self.logger.info("Retrieved group roles list.")
        return RoleList(**res.json())

    @validate_arguments
    async def add_group_users(self, group_id: int, user_list: List[int], raise_on_err: bool = False) -> Group:
        """ bulk add a list of users to a group (by id) """
        if not await self.dracoon.test_connection() and self.dracoon.connection:
            await self.dracoon.connect(OAuth2ConnectionType.refresh_token)

        if self.raise_on_err:
            raise_on_err = True
        
 
        api_url = self.api_url + f'/{str(group_id)}/users'

        payload = {
            "ids": user_list
        }

        try:
            res = await self.dracoon.http.post(url=api_url, json=payload)
            res.raise_for_status()
        except httpx.RequestError as e:
            await self.dracoon.handle_connection_error(e)
        except httpx.HTTPStatusError as e:
            self.logger.error("Adding group users failed.")
            await self.dracoon.handle_http_error(err=e, raise_on_err=raise_on_err)
        
        self.logger.info("Added group user(s).")
        return Group(**res.json())

    @validate_arguments
    async def delete_group_users(self, group_id: int, user_list: List[int], raise_on_err: bool = False) -> Group:
        """ bulk delete a list of users to a group (by id) """
        if not await self.dracoon.test_connection() and self.dracoon.connection:
            await self.dracoon.connect(OAuth2ConnectionType.refresh_token)

        if self.raise_on_err:
            raise_on_err = True

        api_url = self.api_url + f'/{str(group_id)}/users'

        payload = {
            "ids": user_list
        }

        try:
            res = await self.dracoon.http.request(method='DELETE', url=api_url, json=payload, headers=self.dracoon.http.headers)
            res.raise_for_status()
        except httpx.RequestError as e:
            await self.dracoon.handle_connection_error(e)
        except httpx.HTTPStatusError as e:
            self.logger.error("Deleting group users failed.")
            await self.dracoon.handle_http_error(err=e, raise_on_err=raise_on_err)
        
        self.logger.info("Deleted group users(s).")
        return Group(**res.json())